# Copyright (C) 2018 Matthew L. Schwartz - all rights reserved
# Not for redistribution

from SapTrap_Builder import SapTrap_Reference as STR
from SapTrap_Builder import SapTrap_Base as STB
import SapTrap_Text as STT
import tkinter as tk
import tkinter.font as tkf
from tkinter import messagebox
from functools import partial
import sys

class Main_Window:
    """ Main Window (scaffold) and basic flow control and I/O functions. """
    def __init__(self, master):
        self.master = master
        self.main_display_width = 0
        self.controller = None
        # Variables used in user interaction
        self.current_step = tk.IntVar()
        self.current_step.set(1)
        self.last_step = tk.IntVar()
        self.last_step.set(1)
        self.max_step = 4
        self.min_step = 1
        self.edit_site = None
        # Keys to be bound to local functions
        self.bases = ['a', 'c', 'g', 't', '-']
        self.amino_acids = ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M',
                            'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y', '*']
        self.additional_controls = ['z', 'r', 's', 'i'] # bind 'i' here for development, change get_user_response
        self.directional = ['<Left>', '<Right>', '<Up>', '<Down>',
                            '<Shift-Left>', '<Shift-Right>', '<Shift-Up>', '<Shift-Down>',
                            '<Command-Left>', '<Command-Right>', '<Command-Up>', '<Command-Down>',
                            '<Command-A>']
        self.tab_controls = ['<Tab>', '<Shift-Tab>']
        # Setup Functions - setup window, identify steps, bind keys to functions
        self.configure_main_window()
        self.determine_program_steps()
        self.bind_keys()

        # Initialize Panel Elements
        self.error_handler = STT.Error_Handling(master, self)
        self.display_strings = Display_Strings(master, self)
        self.settings_panel = Settings(master, self)
        self.main_display_panel = Main_Display(master, self)
        self.instruction_panel = Instructions(master, self)
        self.progress_panel = Progress_Buttons(master, self)

        # Grid Panel Elements
        self.settings_panel.grid(row=0, column=0, columnspan=3, sticky=tk.NW)
        self.progress_panel.grid(row=0, column=4, sticky=tk.NE + tk.W)
        self.main_display_panel.grid(row=1, column=0, columnspan=5, rowspan=16, sticky=tk.W + tk.E + tk.N + tk.S)
        self.instruction_panel.grid(row=17, column=0, columnspan=5, sticky=tk.W + tk.E)
        self.master.createcommand('tkAboutDialog', self.aboutSapTrap)

    def aboutSapTrap(self):
        messagebox.showinfo(STT.aboutSapTrap['Title'], STT.aboutSapTrap['Text'])

    def configure_main_window(self):
        """ Sets size of main window and window frames based on screen size """
        main_width = int(0.9 * self.master.winfo_screenwidth())
        main_height = int(0.5 * self.master.winfo_screenheight())
        self.master.configure(bg='grey', border=10)
        self.master.title('SapTrap Builder')
        self.master.geometry('{0}x{1}'.format(main_width, main_height))
        font = tkf.Font(family='Courier', size='14')
        label_display_width = 30 * int(font.measure('m'))  # display width of label_display in pixels
        self.main_display_width = int((main_width - label_display_width - 45) /
                                 font.measure('m'))  # max width of main_display in characters
        for x in range(4):
            self.master.columnconfigure(x, weight=1)
        for y in range(18):
            self.master.rowconfigure(y, weight=1)

    def bind_keys(self):
        """ Binds keys to get_user_response method """
        for list in [self.amino_acids, self.bases, self.additional_controls,
                     self.directional, self.tab_controls]:
            for key in list:
                self.master.bind(key, partial(self.get_user_response, key))
            # Note tab keys are also bound to Main_Display.main_display in separate function
        self.master.bind('<Command-s>', self.format_sequence)
        self.master.bind('<Command-u>', self.change_selection_case)
        self.master.bind('<Command-l>', self.change_selection_case)

        return 0

    def determine_program_steps(self):
        """ Determines the range of valid program steps
        This generalizes the GUI, if new step added to program_control dict, no need to update GUI """
        steps = []
        for key in STR.dict_program_control.keys():
            steps.append(int(key))
            steps.sort()
        self.max_step = max(steps)
        self.min_step = min(steps)

        return 0

    def ok_to_change_step(self, response):
        """ This function handles entry into and exit from Sequence entry. Uses current_step.get() and identifier of
            requested next step to determine if step change should proceed"""
        ok_to_change = True
        if self.current_step.get() == 1:
            if response == 'Sequence Entry':
                ok_to_change = False
            else:
                if not self.controller.sequence_entry():
                    ok_to_change = False
        elif response == 'Sequence Entry':
            if self.current_step.get() != 1:
                if not self.controller.sequence_entry_setup():
                    ok_to_change = False
        elif response in ['Review Design', 'Get Reagent Sequences']:
            if not self.controller.final_entry_validation(self.edit_site.final_reagents):
                ok_to_change = False

        return ok_to_change

    def change_active_step(self, response):
        """ input is a program step identifier. Function changes display elements to match identified step.
            Changes both self.current_step and self.last_step appropriately"""
        if self.ok_to_change_step(response):
            self.instruction_panel.update_instructions(response)
            self.last_step.set(self.current_step.get())
            for key,value in STR.dict_program_control.items():
                if value == response:
                    self.current_step.set(int(key))
            for i in range(len(self.progress_panel.progress_buttons)):
                if i == (self.current_step.get()) - 1:
                    self.progress_panel.progress_buttons[i].configure(font=('TkDefaultFont', '13', 'bold'))
                else:
                    self.progress_panel.progress_buttons[i].configure(font=('TkDefaultFont', '13'))
            self.controller.flow_control(response)
        return 0

    def step_forward(self):
        """ Changes current program step by increment of + 1 """
        current_step = self.current_step.get()
        next_step = STB.incrementer(current_step, self.max_step)

        return self.change_active_step(STR.dict_program_control[str(next_step)])

    def step_backward(self):
        """ Changes current program step by increment of - 1 """
        current_step = self.current_step.get()
        next_step = STB.incrementer(current_step, self.max_step, increment=-1, minimum=self.min_step)

        return self.change_active_step(STR.dict_program_control[str(next_step)])

    def get_user_response(self, key, event):
        """ Triggered by key and button presses, sets user_response variable based on event. """
        text = ''
        if key in ['<Left>', '<Right>', '<Up>', '<Down>']:
            text = event.keysym.lower()
        elif key in ['<Shift-Left>', '<Shift-Right>', '<Shift-Up>', '<Shift-Down>']:
            text = event.keysym.upper()
        elif key in ['<Command-Left>', '<Command-Right>', '<Command-Up>', '<Command-Down>']:  # Only on mac; need to update function with windows patch
                text = 'Command-' + event.keysym.lower()
        elif key in ['<Tab>', '<Shift-Tab>']:
            if key == '<Tab>':
                self.step_forward()
                return 'break'
            else:
                return self.step_backward()
        elif key == '<Command-A>':
            text = 'Autoprocess'
        else:
            text = event.char

        return self.controller.flow_control(text)

    def format_sequence(self, event):
        """ Removes non-DNA characters from sequence in the input / main_display box
            Trims lines to length of highlighted region of sequence, or to 100 bases if no highlighting. """
        if self.current_step.get() == 1:
            sequence = self.main_display_panel.main_display.get('1.0', tk.END)
            try:
                length = len(self.main_display_panel.main_display.get(tk.SEL_FIRST, tk.SEL_LAST))
            except:
                length = 100
            sequence = STB.strip_DNA_sequence(sequence, length)
            self.main_display_panel.main_display.delete('1.0', tk.END)
            self.main_display_panel.main_display.replace('1.0', tk.END, sequence)

        return 0

    def change_selection_case(self, event):
        """ Changes the case of selected sequence to uppercase (u) or lowercase (l). """
        if self.current_step.get() == 1:
            try:
                sequence = self.main_display_panel.main_display.get(tk.SEL_FIRST, tk.SEL_LAST)
                change = event.char
                sequence = STB.change_DNA_case(sequence, change)
                self.main_display_panel.main_display.replace(tk.SEL_FIRST, tk.SEL_LAST, sequence)
            except:
                return 0
        return 0


class Settings(tk.Frame):
    """ Contains user options like site name, homology arm lengths, etc. """
    def __init__(self, master, main_window):
        tk.Frame.__init__(self, master)
        self.master = master
        self.main_window = main_window

        # Setup OptionMenus and variables to store selections
        self.jobname = tk.StringVar()
        self.sap_selection = tk.StringVar(master)
        self.output_type = tk.StringVar(master)
        self.sap_options = {'SapTrap', 'Markerless + Insert', 'Markerless No Insert'}
        self.sap_selection.set('SapTrap')
        self.output_options = ['Oligos + Annotated', 'Oligos', 'Synthetic DNA', 'Annotated Synthetic DNA', 'Alignment']
        self.output_type.set('Oligos + Annotated')

        # Header labels for column sets 1 and 2
        self.job_info_label = tk.Label(self, text='Job Information')
        self.job_info_label.grid(row=0, column=0, columnspan=2)
        self.column_spacer = tk.Label(self, text='    ')
        self.column_spacer.grid(row=0, column=2)
        self.homology_arm_label = tk.Label(self, text='Homology Arm Setup')
        self.homology_arm_label.grid(row=0, column=3, columnspan=2)
        # Labels for column set 1
        self.jobname_label = tk.Label(self, text='Job Name:')
        self.jobname_label.grid(row=1, column=0)
        self.sap_selector_label = tk.Label(self, text='  Type:')
        self.sap_selector_label.grid(row=2, column=0)
        self.output_type_label = tk.Label(self, text=' Output:')
        self.output_type_label.grid(row=3, column=0)
        # User entry widgets for column set 1
        self.jobname_entry = tk.Entry(self, width=20, font=('Courier', '14'), textvariable=self.jobname)
        self.jobname_entry.grid(row=1, column=1)
        self.sap_selector = tk.OptionMenu(self, self.sap_selection, *self.sap_options, command=self.apply_sap_option)
        self.sap_selector.grid(row=2, column=1, sticky=tk.E + tk.W)
        self.output_type_selector = tk.OptionMenu(self, self.output_type, *self.output_options)
        self.output_type_selector.grid(row=3, column=1, sticky=tk.E + tk.W)
        # Labels for column set 2
        self.hUS_length_label = tk.Label(self, height = 1, text='hUS target length: ')
        self.hUS_length_label.grid(row=1, column=3)
        self.hDS_length_label = tk.Label(self, height = 1, text='hDS target length: ')
        self.hDS_length_label.grid(row=2, column=3)
        self.min_length_label = tk.Label(self, height=1, text='Min true length:')
        self.min_length_label.grid(row=3, column=3)
        # User entry widgets for column set 2
        self.hUS_length_entry = tk.Text(self, height=1, width=3, font=('Courier', '14'))
        self.hUS_length_entry.grid(row=1, column=4)
        self.hUS_length_entry.insert(tk.END, '57')
        self.hDS_length_entry = tk.Text(self, height=1, width=3, font=('Courier', '14'))
        self.hDS_length_entry.grid(row=2, column=4)
        self.hDS_length_entry.insert(tk.END, '57')
        self.min_length_entry = tk.Text(self, height=1, width=3, font=('Courier', '14'))
        self.min_length_entry.grid(row=3, column=4)
        self.min_length_entry.insert(tk.END, '32')

    def apply_sap_option(self, event):
        """ Calls main_window sap_option setter function when sap_option optionbox is updated. """
        return self.main_window.controller.apply_sap_option()


class Main_Display(tk.Frame):
    """ Main display tk.Entry object. User input of initial sequence, display of sequences throughout design. """
    def __init__(self, master, main_window):
        tk.Frame.__init__(self, master)
        self.master = master
        self.main_window = main_window
        self.show_insert = False
        self.ordered_display_strings = ['base_translation', 'translation_differences', 'mutant_translation',
                                        'pointer', 'mutated_sequence', 'repair_template', 'DNA_differences',
                                        'sequence', 'gRNA', 'PAM', 'Enzyme_labels', 'SapI in guide',
                                        'Terminator in guide', 'Cas9 ends in C', 'intact gRNA site',
                                        'weak gRNA disruption', 'short hUS', 'short hDS',
                                        'long hUS', 'long hDS', 'Trap enzyme in repair template', 'Unlikely point mutants',
                                        'final_display']
        self.config(bg='grey')

        self.label_display = tk.Label(self, width=30, height=13, anchor=tk.NE, bg='grey', justify=tk.RIGHT,
                                      font=('Courier', '14'))
        self.main_display = tk.Text(self, height=13, width=main_window.main_display_width, state='normal', wrap='none',
                               font=('Courier', '14'))
        self.label_display.grid(row=0, column=0, sticky=tk.E + tk.N + tk.S + tk.W)
        self.main_display.grid(row=0, column=1, sticky=tk.W + tk.E + tk.N + tk.S)
        #for key in self.main_window.tab_controls:
         #   self.main_display.bind(key, partial(self.main_window.get_user_response, key))
        self.scrollset()
        self.tab_binding()

    def tab_binding(self):
        for key in self.main_window.tab_controls:
            self.main_display.bind(key, partial(self.main_window.get_user_response, key))

    def get(self):
        """ get text from main_display. """
        return self.main_display.get('1.0', tk.END)

    def scrollset(self):
        """Resets scrollbar after sequence updates. """
        scrollX = tk.Scrollbar(self, orient=tk.HORIZONTAL, command=self.main_display.xview)
        scrollX.grid(row=1, column=1, columnspan=4, sticky=tk.NE + tk.NW)
        self.main_display['xscrollcommand'] = scrollX.set
        scrollY = tk.Scrollbar(self, orient=tk.VERTICAL, command=self.main_display.yview)
        scrollY.grid(row=0, column=2, rowspan=15, sticky=tk.NW + tk.SW)
        self.main_display['yscrollcommand'] = scrollY.set

        return 0

    def update_display(self, location=None, text=None):
        """ Input the text lines to display
            No return, Function updates the displayed text. """
        strings = self.main_window.display_strings.string_library
        #self.silent_site_string()
        main_display_labels = ''
        self.main_display.configure(state='normal')
        self.main_display.delete('1.0', tk.END)
        for tag in self.main_display.tag_names():
            self.main_display.tag_delete(tag)
        line_index = 1
        for identifier in self.ordered_display_strings:
            if strings[identifier].active == True:
                main_display_labels += strings[identifier].label + '\n'
                print_string = strings[identifier].print_string(location=location, text=text,
                                            show_insert=self.main_window.edit_site.show_insert) + '\n'
                self.main_display.insert(tk.END, print_string)
                if strings[identifier].tags != []:
                    for index in range(len(strings[identifier].tags)):
                        tag_name = str(line_index) + '.' + str(index)
                        self.add_tag(tag_name, self.main_display, line_index, strings[identifier].tags[index])
                line_index += print_string.count('\n')
        self.label_display.configure(text=main_display_labels)
        if self.main_window.current_step.get() not in [1, 8]:
            self.main_display.configure(state='disabled')

        return 0

    def add_tag(self, name, text, line, specification):
        """ Name is tag name, text is window, line is line in textbox to tag, and specification is tag specification
            function highlights text or background of specified text. """
        text.tag_add(name, str(line) + specification.start, str(line) + specification.end)
        if specification.highlight == 'foreground':
            text.tag_config(name, foreground=specification.color)
        elif specification.highlight == 'background':
            text.tag_config(name, background=specification.color)

        return 0


class Progress_Buttons(tk.Frame):
    """ Buttons and an option menu corresponding to program steps. """
    def __init__(self, master, main_window):
        tk.Frame.__init__(self, master)
        self.master = master
        self.main_window = main_window
        self.config(bg='grey')
        # Progress Menu - Contains Button widgets outlining the suggested program flow
        self.tag_type = tk.StringVar(master)
        self.tag_types = ['N-tag', 'C-tag', 'Internal - with Linkers', 'Internal, no Linkers', 'Native Intron',
                          'Syntron']
        self.tag_keys = {'N-tag': 'N', 'C-tag': 'C', 'Internal - with Linkers': 'L', 'Internal, no Linkers': 'X',
                    'Native Intron': 'I', 'Syntron': 'S'}
        self.tag_type.set('Select Tag Type')
        self.directional_control = tk.Frame(self)
        self.directional_control.grid(row=4, column=3, columnspan=1, sticky=tk.NE)
        self.progress_buttons = []
        for key in STR.dict_program_control.keys():
            index = int(key) - 1
            row = index % 4
            column = 1 + 2 * int(index / 4)
            if STR.dict_program_control[key] == 'Select Tag Type':
                self.progress_buttons.append(tk.OptionMenu(self, self.tag_type, *self.tag_types,
                                                           command=self.apply_tag_type))
            else:
                self.progress_buttons.append(tk.Button(self, text=STR.dict_program_control[key], width=20,
                                   command=partial(self.main_window.change_active_step, STR.dict_program_control[key])))
            self.progress_buttons[index].grid(row=row, column=column, sticky=tk.E+tk.W)
        self.progress_buttons[0].configure(font=('TkDefaultFont', '13', 'bold'))
        advance_button = tk.Button(self.directional_control, text='>', font=('Courier', '20', 'bold'),
                                   command=self.main_window.step_forward)
        advance_button.grid(row=0, column=1, sticky=tk.E)
        regress_button = tk.Button(self.directional_control, text='<', font=('Courier', '20', 'bold'),
                                   command=self.main_window.step_backward)
        regress_button.grid(row=0, column=0, sticky=tk.E)

    def apply_tag_type(self, event):
        """ Calls function to update tag_type in site.final_reagents when optionbox is changed. """
        self.main_window.controller.apply_tag_type()

class Instructions(tk.Frame):
    """ Panel below main display displays active step specific instructions """
    def __init__(self, master, main_window):
        tk.Frame.__init__(self, master)
        self.main_window = main_window
        self.main_label = tk.Label(self, bg='grey', height=5, width=main_window.main_display_width, justify=tk.LEFT,
                                   anchor=tk.NW)
        self.main_label.grid(row=0, column=1, columnspan=2, sticky=tk.NW)
        self.spacer = tk.Label(self, width=30, height=5, anchor=tk.NE, bg='grey', justify=tk.RIGHT,
                                      font=('Courier', '14'))
        self.spacer.grid(row=0, column=0, sticky=tk.E + tk.N + tk.S + tk.W)
        self.update_instructions('Sequence Entry')

    def update_instructions(self, current_step):
        """ Input is a step name from STR.dict_program_control
            Updates instruction label to match current step """
        text = current_step + '\n' + STT.program_instructions_labels[current_step]
        self.main_label.configure(text=text)

        return 0


class Display_Strings:
    """ Holds and manipulates individual display line item strings. """
    def __init__(self, master, main_window):
        self.master = master
        self.main_window = main_window
        self.string_library = {}
        self.generate_string_library()

    def reset(self):
        """ Resets string library to default configuration. """
        self.string_library = {}
        self.generate_string_library()

        return 0

    def add_display_strings(self, add_strings):
        """ Input is a list of string name keywords. Sets .active property of string to True. """
        for string in add_strings:
            self.string_library[string].active = True

        return 0

    def remove_display_strings(self, remove_strings):
        """ Input is a list of string name keywords. Sets .active property of string to False. """
        if 'repair_template' in remove_strings and self.string_library['DNA_differences'].active == True:
            self.string_library['mutated_sequence'].active = True
        for string in remove_strings:
            self.string_library[string].active = False

        return 0

    def update_active_flags(self, args):
        """ Input is keyword specifying flag string. If specified, .active property changed to True
            otherwise .active property changed to False. """
        for flag in STT.flags.keys():
            if flag in args:
                self.string_library[flag].active = True
            else:
                self.string_library[flag].active = False

        return 0

    def highlight_enzyme(self, enzyme, color):
        """ Input is restriction enzyme name and color for highlighting.
            Adds enzyme and color to list of enzymes in indicated string libraries to be highlighted. """
        for string in [self.string_library['mutated_sequence'], self.string_library['sequence'],
                       self.string_library['repair_template']]:
            if enzyme not in string.highlight_enzymes.keys():
                string.highlight_enzymes[enzyme] = {'name': enzyme, 'color': color, 'locations': []}
        self.label_enzymes()
        return 0

    def reset_enzymes(self):
        """ Deletes all highlighted enzymes from indicated string libraries. """
        for string in [self.string_library['mutated_sequence'], self.string_library['sequence'],
                       self.string_library['repair_template']]:
            string.highlight_enzymes = {}

    def remove_enzyme_highlighting(self, enzyme):
        """ Removes string highlighting from line display items. """
        for string in [self.string_library['mutated_sequence'], self.string_library['sequence'],
                       self.string_library['repair_template']]:
            if enzyme in string.highlight_enzymes:
                del string.highlight_enzymes[enzyme]
            string.clear_tags()
        self.label_enzymes()

    def label_enzymes(self):
        """ Generates a string with names of all restriction enzymes highlighted in indicated string libraries.
            enzyme names are located at position of enzyme site for aligned display. """
        highlight_enzymes = {}
        base_string = ' ' * len(self.string_library['sequence'].base_string)
        print_set = [base_string]
        for string in [self.string_library['mutated_sequence'], self.string_library['sequence']]:
            for enzyme in string.highlight_enzymes.keys():
                if enzyme not in highlight_enzymes.keys():
                    highlight_enzymes[enzyme] = string.highlight_enzymes[enzyme]
        for sequence in [self.string_library['mutated_sequence'], self.string_library['sequence']]:
            for enzyme in highlight_enzymes.keys():
                for location in STB.restriction_site_ranges(enzyme, sequence.base_string):
                    index = 0
                    printed = False
                    while not printed:
                        print_location = print_set[index][location[0]:location[0] + len(enzyme):]
                        if print_location == enzyme:
                            printed = True
                        elif  print_location != ' ' * len(enzyme):
                            if index == len(print_set) - 1:
                                print_set.append(base_string)
                            index += 1
                        else:
                            print_set[index] = print_set[index][:location[0]] + enzyme + \
                                               print_set[index][location[0] + len(enzyme):]
                            printed = True
        print_string = ''
        for item in print_set:
            print_string += item + '\n'

        self.string_library['Enzyme_labels'].base_string = print_string

        return 0


    def generate_string_library(self):
        """ Generate default string library with all possible display line items present and set to .active = False. """
        self.string_library['base_translation'] = Display_String('Wild Type Translation:', 'spacer', 'standard')
        self.string_library['translation_differences'] =  Display_String('', 'spacer', 'standard')
        self.string_library['mutant_translation'] = Display_String('Repair Template Translation:', 'spacer','standard')
        self.string_library['pointer'] = Display_String('', 'spacer', 'standard')
        self.string_library['mutated_sequence'] = Display_String('Repair Template DNA:', 'text', 'standard')
        self.string_library['repair_template'] = Display_String('Repair Template DNA:', 'text', 'standard')
        self.string_library['DNA_differences'] = Display_String('', 'spacer', 'standard')
        self.string_library['sequence'] = Display_String('Wild Type DNA:', 'text', 'standard')
        self.string_library['gRNA'] = Display_String('Guide RNA binding site:', 'intext only', 'standard')
        self.string_library['PAM'] = Display_String('PAM:', 'anchor', 'standard')
        self.string_library['Enzyme_labels'] = Display_String('Restriction Sites:', 'enzyme labels', 'standard')
        self.string_library['final_display'] = Display_String('', 'none', 'standard')
        self.string_library['SapI in guide'] = Display_String('', 'flag', 'flag')
        self.string_library['Terminator in guide'] = Display_String('', 'flag', 'flag')
        self.string_library['Cas9 ends in C'] = Display_String('', 'flag', 'flag')
        self.string_library['intact gRNA site'] = Display_String('', 'flag', 'flag')
        self.string_library['weak gRNA disruption'] = Display_String('', 'flag', 'flag')
        self.string_library['short hUS'] = Display_String('', 'flag', 'flag')
        self.string_library['short hDS'] = Display_String('', 'flag', 'flag')
        self.string_library['long hUS'] = Display_String('', 'flag', 'flag')
        self.string_library['long hDS'] = Display_String('', 'flag', 'flag')
        self.string_library['Trap enzyme in repair template'] = Display_String('', 'flag', 'flag')
        self.string_library['Unlikely point mutants'] = Display_String('', 'flag', 'flag')
        # Add tags to flag elements
        self.string_library['SapI in guide'].add_tag('line', 'red', '.0', '.end')
        self.string_library['Terminator in guide'].add_tag('line', 'red', '.0', '.end')
        self.string_library['Cas9 ends in C'].add_tag('line', 'red', '.0', '.end')
        self.string_library['intact gRNA site'].add_tag('line', 'red', '.0', '.end')
        self.string_library['weak gRNA disruption'].add_tag('line', 'orange', '.0', '.end')
        self.string_library['short hUS'].add_tag('line', 'orange', '.0', '.end')
        self.string_library['short hDS'].add_tag('line', 'orange', '.0', '.end')
        self.string_library['long hUS'].add_tag('line', 'blue', '.0', '.end')
        self.string_library['long hDS'].add_tag('line', 'blue', '.0', '.end')
        self.string_library['Trap enzyme in repair template'].add_tag('line', 'red', '.0', '.end')
        self.string_library['Unlikely point mutants'].add_tag('line', 'red', '.0', '.end')
        self.highlight_enzyme(STR.trap_enzyme, 'salmon')


class Display_String:
    """ Individual display line object. Represents a single line of the display. """
    def __init__(self, label, insert_type, string_type):
        self.label = label
        self.insert = insert_type
        self.string_type = string_type
        self.base_string = ''
        self.pad_length = 0
        self.active = False
        self.highlight_enzymes = {}
        self.tags = []

    def update(self, updated_string, pad_length=0):
        """ Update contents of base_string to print and pad length for printing. """
        self.base_string = str(updated_string)
        self.pad_length = pad_length

        return 0

    def add_tag(self, type, color, start, end):
        """ Input type of sequence, color for highlighting, and start and end of feature to be tagged. """
        self.tags.append(FlagTag(type, color, start, end))

        return 0

    def clear_enzymes(self):
        """ Reset self.highlight_enzymes to empty library. """
        self.highlight_enzymes = {}

        return 0

    def convert_string_to_DNA(self, sequence):
        """ Might be a relic, not sure this is ever called. """
        converted_string = ''
        for char in sequence:
            if char in 'ACGTacgt-':
                converted_string += char
            else:
                converted_string += 'I'
        return STB.DNA(converted_string)

    def clear_tags(self):
        """ Remove highlighting from display string. """
        self.tags = []

    def generate_enzyme_tags(self, enzyme, color, print_sequence):
        """ Input is enzyme name, color to highlight, and print_sequence of the sequence to tag.
            Outputs tag coordinates to generate highlighting of the enzyme site. """
        if '|' in print_sequence:
            insert_location = print_sequence.find('|')
            base_sequence = print_sequence[:insert_location] + print_sequence[insert_location + 10:]
        else:
            insert_location = 0
            base_sequence = print_sequence
        enzyme_site_ranges = STB.restriction_site_ranges(enzyme, base_sequence)
        if insert_location:
            for site_range in enzyme_site_ranges:
                if insert_location in site_range:
                    if insert_location != site_range.start:
                        enzyme_site_ranges.remove(site_range)
            for i in range(len(enzyme_site_ranges)):
                if insert_location < enzyme_site_ranges[i].start + 1:
                    enzyme_site_ranges[i] = range(enzyme_site_ranges[i].start + 10, enzyme_site_ranges[i].stop + 10)
        for site in enzyme_site_ranges:
            start = '.' + str(site.start)
            end = '.' + str(site.stop)
            self.add_tag('enzyme_site', color, start, end)
            self.highlight_enzymes[enzyme]['locations'].append(site.start)

        return 0

    def print_string(self, location=None, text=None, show_insert=False):
        """ Generate string to print to screen - adds insert sequence and generates enzyme tags"""
        if show_insert == True:
            string = self.add_insert_text(location, text)
        else:
            string = (self.pad_for_display(self.pad_length))
        if self.highlight_enzymes != {}:
            self.clear_tags()
            for enzyme in self.highlight_enzymes.keys():
                self.generate_enzyme_tags(self.highlight_enzymes[enzyme]['name'],
                                          self.highlight_enzymes[enzyme]['color'], string)
        return string


    def pad_for_display(self, number, character=' '):
        """ Input is number, produces string of character of len number
        Default character is ' ', but can be changed with optional character argument """
        return character * number + self.base_string

    def add_insert_text(self, location, text):
        """ modifies display text in display_string library by adding |-insert-| text where appropriate. """
        string = self.pad_for_display(self.pad_length)
        if location >= len(string):
            return string
        if self.insert == 'spacer':
            string = string[:location] + '          ' + string[location:]
        elif self.insert == 'text':
            string = string[:location] + text + string[location:]
        elif self.insert == 'intext only':
            if string[location - 1] != ' ' and string[location] != ' ':
                string = string[:location] + text + string[location:]
            else:
                string = string[:location] + '          ' + string[location:]
        elif self.insert == 'enzyme labels':
            strings = []
            if not string.find('\n'):
                strings.append(string)
            else:
                while string.find('\n') != -1:
                    break_location = string.find('\n')
                    strings.append(string[:break_location])
                    string = string[break_location + 1:]
            for i in range(len(strings)):
                while strings[i][location - 1] != ' ':
                    location += 1
                strings[i] = strings[i][:location] + '          ' + strings[i][location:]
                if i < max(range(len(strings))):
                    strings[i] += '\n'
            string = ''
            for line in strings:
                string += line

        elif self.insert == 'anchor':
            anchor_metrics = self.find_anchor(string)
            if anchor_metrics['type'] == 'left':
                if location in anchor_metrics['positions']:
                    location = anchor_metrics['positions'][-1] + 1
            elif anchor_metrics['type'] == 'right':
                if location in anchor_metrics['positions']:
                    location = anchor_metrics['positions'][0] - 1
            string = string[:location] + '          ' + string[location:]
        elif self.insert == 'flag':
            i = 0
            while self.base_string[i] == ' ':
                i += 1
            if i > location:
                string = '          ' + string
        else:
            string = string

        return string

    def find_anchor(self, string):
        """ locates character features in gRNA binding site needed to produce proper insert site display. """
        string_positions = {'open': 0, 'close': 0, 'other': 0}
        protect = 'none'
        for i in range(len(string)):
            if string[i] == '(':
                string_positions['open'] = i
            elif string[i] == ')':
                string_positions['close'] = i
            elif string[i] != ' ':
                if i > string_positions['other']:
                    string_positions['other'] = i
        if string_positions['other'] > string_positions['open']:
            protect = 'left'
        if string_positions['other'] < string_positions['close']:
            protect = 'right'
        protect_positions = range(string_positions['open'] + 1, string_positions['close'] + 1)

        return {'type': protect, 'positions': protect_positions}

class FlagTag:
    """ Holds information for adding tags to individual print lines. """
    def __init__(self, type, color, start, end):
        self.type = type
        self.color = color
        self.highlight = 'foreground'
        self.start = start
        self.end = end
        self.format_tag()


    def format_tag(self):
        """ Determines if tag to be applied to line in display is text or background. """
        if self.type == 'line':
            self.highlight = 'foreground'
        elif self.type == 'enzyme_site':
            self.highlight = 'background'

        return 0
