# Copyright (C) 2018 Matthew L. Schwartz - all rights reserved
# Not for redistribution

import SapTrap_Builder.SapTrap_Reference as STR
import operator

"""
Classes:
    DNA
    GuideSpecification
    GuideRNABinding_Site(DNA)
    EditSite
    FinalReagents

"""


class DNA:

    def __init__(self, sequence):
        self.sequence = sequence
        self.features = {}

    def __len__(self):
        return len(self.sequence)

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < len(self.sequence):
            result = self.sequence[self.n]
            self.n += 1
            return result
        else:
            raise StopIteration

    def __str__(self):
        return self.sequence

    def __add__(self, other):
        return self.sequence + other

    def __getitem__(self, slice_notation):
        return self.sequence[slice_notation]

    def strip_non_DNA(self):
        """ strips any characters that are not DNA bases from sequence """
        return ''.join([base for base in self.sequence if base in STR.dict_degenerate_com.keys()])

    def reverse(self):
        """ returns sequence in reverse order """
        return DNA(self.sequence[::-1])

    def complement(self):
        """generates complement of sequence"""
        return DNA(''.join(STR.dict_degenerate_com[base] for base in self.sequence))

    def reverse_complement(self):
        """generates reverse-complement of sequence"""
        return DNA(''.join(STR.dict_degenerate_com[base] for base in self.sequence)[::-1])

    def mutate(self, position_start, mutated_sequence, position_end=None):
        """ function inserts mutated_sequence string between position_start and position_end in self.sequence
            may be obsolete, replaced by incorporate mutations """
        if position_end == None:
            position_end = position_start
        self.sequence = self.sequence[:position_start] + mutated_sequence + self.sequence[position_end + 1:]
        return 0

    def degenerate_DNA_match(self, seq1, seq2):
        """ Returns True is seq1 == seq2 (degenerate), False if seq1 != seq2 """
        assert len(seq1) == len(seq2)
        for pair in zip(seq1, seq2):
            if not set(STR.degenerate_DNA_code[pair[0]]) & set(STR.degenerate_DNA_code[pair[1]]):
                return False
        return True

    def find_next_instance(self, subject, query):
        """ degernerate search. if seq2 in seq1, returns position of first instance, else returns -1 """
        for i in range(len(subject) - len(query) + 1):
            if self.degenerate_DNA_match(subject[i:i + len(query)], query):
                return i
        return -1

    def find(self, query_sequence, start=0, case_sensitive=False, search_rev_com=False):
        """ Input is query DNA seqeunce
        Output is location of first base of first instance of sequence """
        subject_sequence = self.sequence[start:]
        if case_sensitive == False:
            subject_sequence = subject_sequence.upper()
            query_sequence = query_sequence.upper()
        next_instance = self.find_next_instance(subject_sequence, query_sequence)
        if search_rev_com:
            rc_query_sequence = DNA(query_sequence).reverse_complement()
            next_instance_rc = self.find_next_instance(subject_sequence, rc_query_sequence)

            if next_instance == -1 or next_instance_rc == -1:
                next_instance = max(next_instance, next_instance_rc)
            else:
                next_instance = min(next_instance, next_instance_rc)
        if next_instance != -1:
            next_instance += start
        return next_instance


    def find_all(self, query_sequence, start=0, case_sensitive=False, search_rev_com=True):
        """ Input is string to search for. Output is list of all locations query is found in self.sequence"""
        found_locations = []
        next_instance = self.find(query_sequence, start=start, case_sensitive=case_sensitive,
                                  search_rev_com=search_rev_com)
        while next_instance > -1:
            found_locations.append(next_instance)
            if search_rev_com and (query_sequence != DNA(query_sequence).reverse_complement()):
                if next_instance == self.find(DNA(query_sequence).reverse_complement(), start=next_instance,
                                              case_sensitive=case_sensitive, search_rev_com=False):
                    found_locations.append(next_instance)
            next_instance = self.find(query_sequence, start=next_instance + 1, case_sensitive=case_sensitive,
                                      search_rev_com=search_rev_com)
        return found_locations


    def contains(self, query_sequence, case_sensitive=False, search_rev_com=True):
        """ Input query_sequence (DNA string to search for).
            Output is True if query in self.sequence, False if query not it self.sequence """
        if self.find(query_sequence, case_sensitive=case_sensitive, search_rev_com=search_rev_com) == -1:
            return False
        else:
            return True

    def lower(self):
        """ returns DNA object with lowercase sequence. """
        return DNA(self.sequence.lower())

    def upper(self):
        """ returns DNA object with uppercase sequence """
        return DNA(self.sequence.upper())

    def islower(self):
        """ Returns True if sequence is lowercase, false if not """
        if self.sequence.islower():
            return True
        else:
            return False

    def isupper(self):
        """ returns True if sequence is uppercase, false if not"""
        if self.sequence.isupper():
            return True
        else:
            return False


class GuideSpecification:
    """ * = non-seed base; ! = seed base, / = top cutsite, \ = bottom cutsite, PAM in basses"""

    def __init__(self, specification_string, name, print_mask):
        self.spec_string = specification_string
        self.print_mask = print_mask
        self.name = name
        self.search_string = ''
        self.non_seed_positions = []
        self.seed_positions = []
        self.PAM_positions = []
        self.cutsite = None
        self.PAM = ''
        self.rc_PAM_offset = 0
        self.parse_spec_string()

    def parse_spec_string(self):
        """ Input is specification string for gRNA. (see SapTrap_Reference for description.
            Reads spec string code and populates guide specification parameters such as PAM_positions, cutsite. """
        sense_cutsite = self.spec_string.find('/')
        self.spec_string = self.spec_string[:sense_cutsite] + self.spec_string[sense_cutsite + 1:]
        antisense_cutsite = self.spec_string.find('\\')
        self.spec_string = self.spec_string[:antisense_cutsite] + self.spec_string[antisense_cutsite + 1:]
        if sense_cutsite != antisense_cutsite:
            self.cutsite = int((sense_cutsite + antisense_cutsite) / 2)
        else:
            self.cutsite = sense_cutsite
        for position in self.spec_string:
            if position in STR.degenerate_DNA_code.keys():
                self.search_string += position
            elif position in '*!':
                self.search_string += 'N'
            elif position == '-':
                self.rc_PAM_offset += 1
        for i in range(len(self.spec_string)):
            if self.spec_string[i] == '*':
                self.non_seed_positions.append(i)
            elif self.spec_string[i] == '!':
                self.seed_positions.append(i)
            elif self.spec_string[i] in STR.degenerate_DNA_code.keys():
                self.PAM_positions.append(i)
        for position in self.PAM_positions:
            self.PAM += self.spec_string[position]

        return 0

    def evaluate_offtarget(self, binding_site, offtarget_site, insertion_site=None):
        """ Input is sequence of gRNA binding site (genomic) and offtarget_site (here, mutated repair template
            over same region as binding site). Optional input position where insert is located within binding site.
            Output is a keyword score indicating the degree of binding site disruption in offtarget_site. """
        disruption_score = {'weak': 0, 'strong': 0}
        if insertion_site:
            if insertion_site in self.non_seed_positions:
                disruption_score['weak'] += 1
            elif insertion_site in self.seed_positions:
                disruption_score['strong'] += 1
            elif insertion_site in self.PAM_positions:
                disruption_score['strong'] += 1
        for i in range(len(binding_site)):
            if i in self.seed_positions:
                if binding_site[i] != offtarget_site[i]:
                    disruption_score['strong'] += 1
            elif i in self.non_seed_positions:
                if binding_site[i] != offtarget_site[i]:
                    disruption_score['weak'] += 1
            elif i in self.PAM_positions:
                PAM_position_set = set(STR.degenerate_DNA_code[self.spec_string[i]])
                if offtarget_site[i].upper() not in PAM_position_set:
                    disruption_score['strong'] += 1
        if disruption_score['strong'] == 0:
            if disruption_score['weak'] == 0:
                return 'none'
            else:
                return 'weak'
        else:
            return 'strong'


class GuideRNABinding_Site(DNA):
    """ This object stores individual guide RNA binding sites, based on GuideSpecification string. """

    def __init__(self, sequence, start, direction, guide_specification):
        DNA.__init__(self, sequence)
        self.start = start
        self.end = start + len(guide_specification.spec_string)
        self.direction = direction
        self.guide_specification = guide_specification
        self.display_sequence = {'sequence': '', 'PAM': ''}
        self.cutsite = 0
        self.PAM = 0
        self.PAM_sequence = None
        self.distance = 0
        self.guide_flags = []
        self.gRNA_sequence = self.calculate_gRNA_sequence()
        self.calculate_cutsite_and_PAM()
        self.flag_guide()

    def calculate_cutsite_and_PAM(self):
        """ Determines sequence of PAM and core binding site from parent object. Runs in __init__()"""
        for i in self.guide_specification.PAM_positions:
            self.display_sequence['PAM'] += self.sequence[i]
        if self.direction == 'sense':
            self.cutsite = self.start + self.guide_specification.cutsite
            self.PAM = self.start + self.guide_specification.PAM_positions[-1]
            self.display_sequence['sequence'] = str(self.sequence)

        elif self.direction == 'antisense':
            self.cutsite = self.start + (len(self.guide_specification.spec_string) - self.guide_specification.cutsite)
            self.PAM = self.start + len(self.guide_specification.spec_string) - \
                       self.guide_specification.PAM_positions[-1]
            self.display_sequence['sequence'] = str(self.reverse_complement())
            self.display_sequence['PAM'] = str(DNA(self.display_sequence['PAM']).complement())

        return 0

    def calculate_gRNA_sequence(self):
        gRNA_sequence = ''
        for i in range(len(self.guide_specification.spec_string)):
            if self.guide_specification.spec_string[i] in ['!', '*']:
                gRNA_sequence += self.sequence[i]

        return gRNA_sequence

    def flag_guide(self):
        """ Checks guide for flags, appends any found flags to objects self.flags. """
        if self.find(STR.restriction_enzyme_list[STR.trap_enzyme], search_rev_com=True) != -1:
            self.guide_flags.append('SapI in guide')
        if self.find('TTTTTT', search_rev_com=False) != -1:
            self.guide_flags.append('Terminator in guide')
        if self.guide_specification.name == 'Cas9':
            if self.sequence[19].lower() == 'c':
                self.guide_flags.append('Cas9 ends in C')

        return 0


class EditSite:
    def __init__(self, sequence, guide_specification, insert_site=None):
        self.reset(sequence, guide_specification, insert_site=insert_site)

    def reset(self, sequence, guide_specification, insert_site=None):
        self.sequence = DNA(sequence)
        self.mutated_sequence = DNA(sequence)
        self.temporary_sequence_hold = DNA(sequence)
        self.guide_specification = guide_specification
        self.insert_site = insert_site
        self.original_insert_site = insert_site
        self.show_insert = False
        self.frame = self.guess_frame()
        self.guide_list = self.sort_guide_list(self.find_guide_sites())
        self.gRNA = None
        self.mutation_site = 0
        self.silent_site_search_region = None
        self.silent_sites = []
        self.silent_site_index = None
        self.final_reagents = FinalReagents(self)

    def guess_frame(self):
        """ Sets the initial translation frame so that insertion site is in-frame
            Should always be correct for in-coding region inserts, can guess wrong for intron / syntron tags"""
        coding_base_count = 0
        for base in self.sequence[: self.insert_site]:
            if base.isupper():
                coding_base_count += 1
        frame = coding_base_count % 3
        return frame

    def incorporate_large_deletion(self, deletion_region):
        """ Called during import if <..> used to indicate user wants to delete a region of the genome
            replaces bases enclosed by <..> with -'s in mutated_sequence """
        deletion_string = self.mutated_sequence[:deletion_region[0]]
        for i in range(deletion_region[0], deletion_region[1]):
            deletion_string += '-'
        self.large_deletion_range = [deletion_region[0], deletion_region[1]]
        deletion_string += self.mutated_sequence[deletion_region[1]:]
        self.mutated_sequence = deletion_string

        return 0

    def get_guide_index(self, gRNA_sequence):
        """ Used for processing multiple guide RNAs - input is gRNA binding site, returns position in
        edit sites guide_list."""
        for key, value in self.guide_list.items():
            if value[:20] == gRNA_sequence:
                return key

        return 1

    def collect_flags(self, program_step):
        flags = []
        flag_sets = []
        if self.gRNA is not None:
            flag_sets.append(self.gRNA.guide_flags)
        if program_step in [7, 8]:
            flag_sets.append(self.final_reagents.flags)
        for flag_set in flag_sets:
            for flag in flag_set:
                if flag not in flags:
                    flags.append(flag)
        if self.final_reagents.sap_option in ['Markerless No Insert', 'Markerless + Insert']:
            for flag in ['SapI in guide', 'SapI in repair template']:
                if flag in flags:
                    flags.remove(flag)

        return flags

    def change_frame(self, increment):
        """ Input increment (1 or -1) sets self.frame to 0, 1, or 2 based on increment plus set value. """
        new_frame = (self.frame + increment) % 3
        self.frame = new_frame
        return new_frame

    def incorporate_mutation(self, position, mutation):
        """ Simply changes the base at position in sequence to mutation within self.mutated_sequence attribute. """
        self.mutated_sequence = DNA(self.mutated_sequence[: position] + mutation +
                                    self.mutated_sequence[position + 1:])

        return 0

    def set_pointer(self, position, marker='|'):
        """ Generates pointer string for display. Sets self.pointer_string, and returns string. """
        output = ''
        for i in range(len(self.sequence)):
            if i == position:
                output += marker
            else:
                output += ' '
        return output

    def translate(self, sequence):
        # Call to translate uppercase, return non-condensed translation for printing
        return translate_upper_case(sequence, self.frame, condensed=False)

    def find_guide_sites(self):
        # input self, returns list of GuideRNABinding objects for all sites in self.sequence
        guide_sites = []
        guide_length = len(self.guide_specification.search_string)
        sense_locations = self.sequence.find_all(self.guide_specification.search_string, search_rev_com=False)
        antisense_locations = self.sequence.find_all(DNA(self.guide_specification.search_string).reverse_complement(),
                                                     search_rev_com=False)
        for location in sense_locations:
            guide_sites.append(GuideRNABinding_Site(self.sequence[location:location + guide_length], location, 'sense',
                                                    self.guide_specification))
        for location in antisense_locations:
            guide_sites.append(GuideRNABinding_Site(str(DNA(self.sequence[location:location +
                                                                                   guide_length]).reverse_complement()),
                                                    location, 'antisense',
                                                    self.guide_specification))
        return guide_sites

    def sort_guide_list(self, guide_list):
        """This function will take a list of guides and assign keys in order of distance between cut and insert"""
        initial_sort = sorted(guide_list, key=operator.attrgetter('cutsite'))
        sorted_guide_list = {}
        count = 1
        for gRNA in initial_sort:
            sorted_guide_list[count] = gRNA
            gRNA.index = count
            count += 1

        return sorted_guide_list

    def sort_guides_by_distance(self):
        """ Used by automatic script, sorts guides by distance from cutsite rather than by linear position
            Requires already populated and sorted guide list in self.guide_list """
        self.calculate_guide_distances()
        initial_sort = sorted(self.guide_list.values(), key=operator.attrgetter('distance'))
        count = 1
        sorted_guide_list = {}
        for gRNA in initial_sort:
            sorted_guide_list[count] = gRNA
            gRNA.index = count
            count += 1

        self.guide_list = sorted_guide_list

    def PAM_string(self):
        """ Generates PAM string for current selected guide (Handles both Cas9 and Cpf1)
            Outputs PAM string for display (PAM, /\ to mark cutsite, and insert-to-cutsite distance
        """
        print_string = ''
        print_mask = self.gRNA.guide_specification.print_mask
        PAM = self.gRNA.display_sequence['PAM']
        distance = '(' + str(abs(self.insert_site - self.gRNA.cutsite)) + ')'
        space_offset = 0
        for char in print_mask:
            if char == '#':
                print_string += distance
                space_offset = len(str(distance))
            elif char == ' ':
                if space_offset > 0:
                    space_offset -= 1
                else:
                    print_string += ' '
            elif char == 'P':
                print_string += PAM
            else:
                print_string += char
        if self.gRNA.direction == 'antisense':
            rc_print_string = ''
            for char in print_string[::-1]:
                if space_offset > 0:
                    space_offset -= 1
                elif char == ')':
                    space_offset = len(str(distance))
                    rc_print_string += distance
                elif char == '/':
                    rc_print_string += '\\'
                elif char == '\\':
                    rc_print_string += '/'
                else:
                    rc_print_string += char
            print_string = rc_print_string

        return print_string

    def calculate_guide_distances(self):
        for i in range(1, 1 + len(self.guide_list)):
            self.guide_list[i].distance = abs(self.guide_list[i].cutsite - self.insert_site)

        return 0

    def closest_cutting_guide(self):
        selection = 1
        self.calculate_guide_distances()
        for i in range(1, 1 + len(self.guide_list)):
            if self.guide_list[i].distance < self.guide_list[selection].distance:
                selection = i
        return selection

    def set_gRNA(self, index):
        """ Input is number of gRNA in gRNA list
            sets self.gRNA to guide object at index """
        self.gRNA = self.guide_list[index]
        return 0

    def insert_text(self):
        """ Determines if current self.insert_site is at a codon boundary(in frame). Returns capital vs lowercase
            text for displaying insert. """
        if self.guess_frame() == self.frame:
            return '|-INSERT-|'
        else:
            return '|-insert-|'

    def feature_locator(self, feature):
        """ input is a keyword string: insert_site, guide, or mutation. Returns string '1.x', x is location
            of referenced feature. This is useful for using the See function to scroll to the feature after
            display updates. """
        output = '1.'
        if feature == 'insert_site':
            output += str(self.insert_site)
        if feature == 'guide':
            if self.gRNA is not None:
                output += str(self.gRNA.cutsite)
            else:
                output += str(self.insert_site)
        if feature == 'mutation':
            output += str(self.mutation_site)
        return output

    def difference_string(self, type, match=' ', mismatch='*'):
        """ Input type ('DNA' or 'translation')
            Output string with match character where sequences match, mismatch character where mismatched """
        output = ''
        string_1 = self.sequence
        string_2 = self.mutated_sequence
        if type == 'translation':
            string_1 = self.translate(string_1)
            string_2 = self.translate(string_2)
        for i in range(len(string_1)):
            try:
                if string_1[i] == string_2[i]:
                    output += match
                else:
                    output += mismatch
            except:
                output += ''

        if self.final_reagents.sap_option == 'SapTrap' and \
                self.final_reagents.tag_type in ['I', 'S'] and type == 'DNA':
            output = self.incorporation_likelihood_string(output, match, mismatch)

        return output

    def incorporation_likelihood_string(self, difference_string, match, mismatch, alt_mismatch='x'):
        """ When called, rewrites a difference string so that positions outside of the insert site and cutsite
            are highlighted by a different character. """
        output = ''
        if self.gRNA is None:
            return difference_string
        range_of_incorporation = range(min(self.insert_site, self.gRNA.cutsite),
                                       max(self.insert_site, self.gRNA.cutsite))
        for i in range(len(difference_string)):
            if difference_string[i] == match:
                output += match
            elif difference_string[i] == mismatch:
                if i in range_of_incorporation:
                    output += mismatch
                else:
                    output += alt_mismatch

        return output

    def get_codon_positions_list(self):
        """ Returns a list of lists. Each list is the positions (numerical string coordinates) of the 3 bases
            comprising a single codon. The main list is a list of all codon position lists. """
        coding_bases = []
        for i in range(len(self.mutated_sequence)):
            if self.mutated_sequence[i].isupper():
                coding_bases.append(i)
        coding_bases = coding_bases[self.frame:]
        coding_bases = coding_bases[: len(coding_bases) - (len(coding_bases) % 3)]
        return list_to_triplets(coding_bases)

    def get_current_codon(self, position):
        """ Input = position in self.mutated_sequence. Locates the codon the position is part of.
            Returns a library with the positions of the entire codon, the codon, and the amino acid. """
        if self.mutated_sequence[position] in 'ACGT':
            codons = self.get_codon_positions_list()
            for codon in codons:
                if position in codon:
                    codon_index = codons.index(codon)
            codon = ''
            try:
                for i in codons[codon_index]:
                    codon += self.mutated_sequence[i]
            except:
                return {}
            amino_acid = translate(codon)
            output_library = {'positions_list': codons[codon_index], 'codon': codon, 'amino_acid': amino_acid}
        else:
            output_library = {}

        return output_library

    def reset_mutator(self, reset_code):
        """ When mutator is entered, a copy of self.mutated_sequence is stored. 'r' resets mutated_sequence to this
            entry snapshot; z resets mutated_sequence to WT (undoes all changes from any session). """
        if reset_code == 'r':
            self.mutated_sequence = self.temporary_sequence_hold
        elif reset_code == 'z':
            self.mutated_sequence = self.sequence
        return 0

    def move_pointer(self, move_code):
        """ input is left or right arrow key press, increments self.mutation_site by 1, 9, -1, or -9. """
        if move_code == 'right':
            self.mutation_site = incrementer(self.mutation_site, len(self.sequence) - 1)
        elif move_code == 'left':
            self.mutation_site = incrementer(self.mutation_site, len(self.sequence), increment=-1)
        elif move_code == 'RIGHT':
            self.mutation_site = incrementer(self.mutation_site,
                                             len(self.sequence) - 1, increment=9)
        elif move_code == 'LEFT':
            self.mutation_site = incrementer(self.mutation_site,
                                             len(self.sequence) - 1, increment=-9)
        return 0

    def trim_starting_ATG(self):
        """ If self.insert_site is boardered by 'ATG', deletes and replaces with '---'
            Only call for N-tag saptrap design, as the 'ATG' is added back as a sticky end.
            without trimming constructs would have 'ATGATG' at start site. """
        lower_bound = None
        upper_bound = None
        if self.mutated_sequence[self.insert_site:self.insert_site + 3] == 'ATG':
            lower_bound = self.insert_site
            upper_bound = self.insert_site + 3
        elif self.mutated_sequence[self.insert_site - 3:self.insert_site] == 'ATG':
            lower_bound = self.insert_site - 3
            upper_bound = self.insert_site
        if lower_bound is not None:
            for i in range(lower_bound, upper_bound):
                self.incorporate_mutation(i, '-')

    def base_mutator(self, base_code):
        """ Input is a base ( 'a', 'c', 'g', 't', or '-')
            changes base at active position to input base """
        if self.mutated_sequence[self.mutation_site] == '-':
            if base_code in 'acgt':
                if self.sequence[self.mutation_site].isupper():
                    self.incorporate_mutation(self.mutation_site, base_code.upper())
                elif self.sequence[self.mutation_site].islower():
                    self.incorporate_mutation(self.mutation_site, base_code.lower())
        elif self.mutated_sequence[self.mutation_site] in 'acgt-':
            self.incorporate_mutation(self.mutation_site, base_code)
        elif self.mutated_sequence[self.mutation_site] in 'ACGT':
            self.incorporate_mutation(self.mutation_site, base_code.upper())

        return 0

    def mutate_non_coding(self, mutation_code):
        """ Not completely sure this is used, input is base to mutate position to
            Output changed mutated_sequence at mutation_site to base. """
        if self.mutated_sequence[self.mutation_site] == '-':
            if mutation_code in 'acgt':
                if self.sequence[self.mutation_site].isupper():
                    self.incorporate_mutation(self.mutation_site, mutation_code.upper())
                elif self.sequence[self.mutation_site].islower():
                    self.incorporate_mutation(self.mutation_site, mutation_code.lower())
        elif self.mutated_sequence[self.mutation_site] in 'acgt':
            self.incorporate_mutation(self.mutation_site, mutation_code)

        return 0

    def amino_acid_mutator(self, amino_acid_code, amino_acid_index=0):
        """ For coding bases, input is single letter amino acid code.
            Function changes DNA so that codon of current base encodes for entered amino acid. """
        if self.sequence[self.mutation_site] in 'ACGT':
            current_codon_information = self.get_current_codon(self.mutation_site)
            current_codon = current_codon_information['codon']
            current_positions = current_codon_information['positions_list']
            if current_codon_information != {}:
                new_codon = synonymous_codons(amino_acid_code)[amino_acid_index]
                for i in range(3):
                    if current_codon[i] == new_codon[i]:
                        continue
                    else:
                        self.incorporate_mutation(current_positions[i], new_codon[i])

        return 0

    def scroll_mutator(self, scroll_command):
        """ Input is a scroll command (up, down, shift-up, shift-down)
            Output changes current base codon to new synonymous or non-synonymous codon. """
        current_codon_information = self.get_current_codon(self.mutation_site)
        if current_codon_information != {}:
            current_codon = current_codon_information['codon']
            current_amino_acid = current_codon_information['amino_acid']
            current_amino_acid_index = STR.amino_acids.index(current_codon_information['amino_acid'])
            silent_mutations = synonymous_codons(current_codon_information['amino_acid'])
            current_codon_index = silent_mutations.index(current_codon)
            if scroll_command in ['up', 'down']:
                if scroll_command == 'up':
                    new_index = incrementer(current_codon_index, len(silent_mutations) - 1, increment=-1)
                elif scroll_command == 'down':
                    new_index = incrementer(current_codon_index, len(silent_mutations) - 1)
                return self.amino_acid_mutator(current_amino_acid, new_index)
            elif scroll_command in ['UP', 'DOWN']:
                if scroll_command == 'UP':
                    new_amino_acid = STR.amino_acids[incrementer(current_amino_acid_index,
                                                                 len(STR.amino_acids) - 1, circular=True)]
                elif scroll_command == 'DOWN':
                    new_amino_acid = STR.amino_acids[incrementer(current_amino_acid_index,
                                                                 len(STR.amino_acids) - 1,
                                                                 increment=-1, circular=True)]
                return self.amino_acid_mutator(new_amino_acid)
        else:
            return 0

    def get_silent_search_region(self, location):

        if self.mutated_sequence[location] in 'acgt-':
            return []
        position = location
        stripped_sequence = strip_deletions(self.mutated_sequence)
        sequence = stripped_sequence['sequence']
        deletions = stripped_sequence['deletions']
        for location in deletions:
            if location < position:
                position -= 1
        codons = codon_positions(sequence, self.frame)
        current_codon_index = None
        current_codon = []
        prior_codon = []
        next_codon = []
        for index in range(len(codons)):
            if position in codons[index]:
                current_codon_index = index
                current_codon = codons[current_codon_index]
        if current_codon_index is None:
            return []
        else:
            if current_codon_index > 0:
                prior_codon = codons[current_codon_index - 1]
            if current_codon_index < len(codons) - 1:
                next_codon = codons[current_codon_index + 1]
        codon_sub_position = codons[current_codon_index].index(position)
        search_max = position + 3 + (3 - codon_sub_position)
        if search_max > len(sequence):
            search_max = len(sequence)
        search_min = position - 3 - (codon_sub_position)
        if search_min < 0:
            search_min = 0
        local_region = range(search_min, search_max)
        output = []
        for codon in [prior_codon, next_codon, current_codon]:
            inscore = 0
            for position in codon:
                if position in local_region:
                    inscore += 1
            if inscore == 3:
                output += codon
        output.sort()
        for deleted_position in deletions:
            for i in range(len(output)):
                if deleted_position <= output[i]:
                    output[i] += 1
        return output

    def find_silent_sites(self, wt_sequence, permutation_array):

        silent_site_list = []
        permutations = []
        for permutation in permutation_array:
            permutations.append(DNA(permutation))
        for enzyme, rec_seq in STR.restriction_enzyme_list.items():
            fewest_changes = ''
            for permutation in permutations:
                if permutation.find(STR.restriction_enzyme_list[enzyme], search_rev_com=True) != -1:
                    if difference_count(wt_sequence, str(permutation)) < \
                            difference_count(wt_sequence, fewest_changes):
                        fewest_changes = str(permutation)
            if fewest_changes != '':
                if fewest_changes != wt_sequence:
                    silent_site_list.append([fewest_changes, enzyme])
        if silent_site_list != []:
            silent_site_list.append([wt_sequence, 'wild_type'])

        return silent_site_list

    def position_list_to_sequence(self, position_list, reference_sequence):
        """ input is a list of positions (integers) in reference_sequence
            output is string comprising the concatenated characters from reference string at positions listed """
        DNA_sequence = ''
        for position in position_list:
            DNA_sequence += reference_sequence[position]
        return DNA_sequence

    def reset_silent_sites(self, region):
        DNA_sequence = self.position_list_to_sequence(region, self.mutated_sequence)
        permutation_array = make_permutation_array(translate(DNA_sequence),
                                              STR.dict_rev_genetic_code_elegans)
        self.silent_sites = self.find_silent_sites(DNA_sequence, permutation_array)
        self.silent_site_search_region = region
        current_sequence = self.position_list_to_sequence(region, self.mutated_sequence)
        self.silent_site_index = 0
        for i in range(len(self.silent_sites)):
            if current_sequence == self.silent_sites[i][0]:
                self.silent_site_index = i

        return 0

    def silent_mutator(self, scroll_command):

        if scroll_command in ['Command-up', 'Command-down']:
            increment = 1
            if scroll_command == 'Command-down':
                increment = -1
            region = self.get_silent_search_region(self.mutation_site)
            if region == []:
                return 0
            if region != self.silent_site_search_region:
                self.reset_silent_sites(region)
            if self.silent_sites == []:
                return 0
            else:
                self.silent_site_index = incrementer(self.silent_site_index, len(self.silent_sites) - 1,
                                                     circular=True, increment=increment)
                new_sequence = self.silent_sites[self.silent_site_index][0]
                for i in range(len(region)):
                    self.incorporate_mutation(region[i], new_sequence[i])

            return self.silent_sites[self.silent_site_index][1]

    def search_for_mutable_sites(self, scroll_command):
        """ from self.mutation_site, searches right or left in increments of 3 bases for next adjacent position
            for which silent restriction sites can be encoded. sets self.mutation_site to next location found. """
        location = self.mutation_site
        location_range = range(len(self.mutated_sequence) - 3)
        silent_sites = []
        increment = 3
        if scroll_command == 'Command-left':
            increment *= -1
        while silent_sites == [] and location in location_range:
            location += increment
            search_sequence = self.position_list_to_sequence(self.get_silent_search_region(location),
                                                             self.mutated_sequence)
            permutation_array = make_permutation_array(translate(search_sequence), STR.dict_rev_genetic_code_elegans)
            silent_sites = self.find_silent_sites(search_sequence, permutation_array)
        if silent_sites != []:
            self.mutation_site = location
        return 0


class FinalReagents:
    def __init__(self, edit_site):
        self.site = edit_site
        self.display_string = ''
        self.display_offset = 0
        self.flags = []
        # Values obtained from Settings and Progress Panels
        self.jobname = ''
        self.sap_option = 'SapTrap'
        self.output_type = 'Oligos + Annotated'
        self.tag_type = None
        self.hUS_target = 57
        self.hDS_target = 57
        self.min_true_length = 32

    def reset_flags(self):
        self.flags = []
        return 0

    def check_min_length(self):

        cutsite = self.site.gRNA.cutsite
        insert_site = self.site.insert_site
        hUS_target = self.hUS_target
        hDS_target = self.hDS_target
        if cutsite < insert_site:
            cut_to_insert = count_bases(self.site.mutated_sequence[cutsite: insert_site])
            true_hUS_length = hUS_target - cut_to_insert
            if true_hUS_length < self.min_true_length:
                hUS_target += self.min_true_length - (hUS_target - cut_to_insert)
        elif cutsite > insert_site:
            insert_to_cut = count_bases(self.site.mutated_sequence[insert_site: cutsite])
            true_hDS_length = hDS_target - insert_to_cut
            if true_hDS_length < self.min_true_length:
                hDS_target += self.min_true_length - (hDS_target - insert_to_cut)

        return {'hUS': hUS_target, 'hDS': hDS_target}

    def get_homology_arms(self):
        if self.tag_type == 'N' and self.sap_option == 'SapTrap':
            self.site.trim_starting_ATG()
        HA_target_lengths = self.check_min_length()
        hUS = get_n_bases(self.site.mutated_sequence, HA_target_lengths['hUS'], self.site.insert_site - 1,
                          direction='left')
        hDS = get_n_bases(self.site.mutated_sequence, HA_target_lengths['hDS'], self.site.insert_site)
        hUS_display = self.site.mutated_sequence[self.site.insert_site - hUS['count']:self.site.insert_site]
        hDS_display = self.site.mutated_sequence[self.site.insert_site:self.site.insert_site + hDS['count']]
        self.display_offset = self.site.insert_site - hUS['count']
        self.display_string = hUS_display + hDS_display
        if len(hUS['sequence']) < self.hUS_target:
            self.flags.append('short hUS')
        if len(hUS['sequence']) > self.hUS_target:
            self.flags.append('long hUS')
        if len(hDS['sequence']) < self.hDS_target:
            self.flags.append('short hDS')
        if len(hDS['sequence']) > self.hDS_target:
            self.flags.append('long hDS')
        for homology_arm in [hUS, hDS]:
            search_arm = DNA(homology_arm['sequence'])
            if search_arm.find(STR.restriction_enzyme_list[STR.trap_enzyme], search_rev_com=True) != -1:
                self.flags.append('Trap enzyme in repair template')
        differences = str(self.site.difference_string('DNA'))
        if 'x' in differences:
            self.flags.append('Unlikely point mutants')
        return {'hUS': hUS, 'hDS': hDS}

    def evaluate_guide_binding_site(self):

        insert_site = None
        guide_sequence = self.site.gRNA.sequence
        binding_site = DNA(self.site.mutated_sequence[self.site.gRNA.start:self.site.gRNA.end])
        if self.site.insert_site in range(self.site.gRNA.start, self.site.gRNA.end):
            insert_site = self.site.insert_site - self.site.gRNA.start
        if self.site.gRNA.direction == 'antisense':
            binding_site = binding_site.reverse_complement()
            if insert_site:
                insert_site = len(binding_site) - insert_site
        if '-' in binding_site:
            disruption_score = self.evaluate_deletion_site()
        else:
            disruption_score = self.site.gRNA.guide_specification.evaluate_offtarget(guide_sequence, binding_site,
                                                                                     insert_site)
        if disruption_score == 'none':
            self.flags.append('intact gRNA site')
        elif disruption_score == 'weak':
            self.flags.append('weak gRNA disruption')

        return 0

    def evaluate_deletion_site(self):
        guide_sequence = self.site.gRNA.sequence
        gRNA_length = len(self.site.gRNA.guide_specification.spec_string)
        deletion_positions = []
        deletion_free_sequence = ''
        insert_site = self.site.insert_site
        gRNA_start = self.site.gRNA.start
        gRNA_end = self.site.gRNA.end
        for i in range(len(self.site.mutated_sequence)):
            if self.site.mutated_sequence[i] != '-':
                deletion_free_sequence += self.site.mutated_sequence[i]
            else:
                deletion_positions.append(i)
        for position in deletion_positions[::-1]:
            if insert_site > position:
                insert_site -= 1
            if gRNA_start > position:
                gRNA_start -= 1
            if gRNA_end > position:
                gRNA_end -= 1
        binding_sites = []
        from_start_insert_site = None
        from_end_insert_site = None
        found_site_insert_site = None
        if (insert_site - gRNA_start) in range(gRNA_length):
            from_start_insert_site = insert_site - gRNA_start
        if (insert_site - (gRNA_end - gRNA_length)) in range(gRNA_length):
            from_end_insert_site = (insert_site - (gRNA_end - gRNA_length))
        binding_sites.append([deletion_free_sequence[gRNA_start:gRNA_start + gRNA_length], from_start_insert_site])
        binding_sites.append([deletion_free_sequence[gRNA_end - gRNA_length:gRNA_end], from_end_insert_site])
        if DNA(deletion_free_sequence).find(guide_sequence, search_rev_com=True):
            found_site = DNA(deletion_free_sequence).find(guide_sequence, search_rev_com=True)
            if insert_site in range(found_site, found_site + gRNA_length):
                found_site_insert_site = insert_site - found_site
                binding_sites.append([deletion_free_sequence[found_site:found_site + gRNA_length],
                                      found_site_insert_site])
        if self.site.gRNA.direction == 'antisense':
            for site in binding_sites:
                site[0] = DNA(site[0]).reverse_complement()
                if site[1]:
                    site[1] = len(site[0]) - site[1]
        disruption_score = 'strong'
        for site in binding_sites:
            score = self.site.gRNA.guide_specification.evaluate_offtarget(guide_sequence, site[0], site[1])
            if score == 'none':
                disruption_score = 'none'
            elif score == 'weak':
                if disruption_score == 'strong':
                    disruption_score = 'weak'
        return (disruption_score)

    def print_oligos(self, view):
        """ Input is string specifying the desired format for the output.
            Returns block of text containing reagent sequences in desired format for user"""
        arm_sequences = self.get_homology_arms()
        hUS = DNA(arm_sequences['hUS']['sequence'])
        hDS = DNA(arm_sequences['hDS']['sequence'])
        output = '\n'

        if view == 'Batch':
            output = {'guide_sequence': self.site.gRNA.gRNA_sequence,
                      'mutated_sequence': self.site.mutated_sequence,
                      'cutsite': self.site.gRNA.cutsite,
                      'insert_site': self.site.insert_site}
            direction = 1
            if self.site.gRNA.cutsite < self.site.insert_site:
                direction = -1
            output['direction'] = direction
            # 'GCGgGAAGAGCgtggtgGCTCTTCgGGT'

        if 'Markerless' in self.sap_option:
            output += 'Guide RNA sequence:\n'
            output += self.site.gRNA.gRNA_sequence + '\n\n'
            output += 'Repair Template Sense Strand\n'
            if self.site.gRNA.direction == 'antisense':
                output += '(this strand can be cleaved by Cas9 + selected gRNA)\n'
            output += str(hUS) + str(hDS) + '\n\n'
            output += 'Repair Template Antisense Strand\n'
            if self.site.gRNA.direction == 'sense':
                output += '(this strand can be cleaved by Cas9 + selected gRNA)\n'
            output += str(hDS.reverse_complement()) + str(hUS.reverse_complement())
            output += '\n\nIndividual Homology Arm sequences:\n'
            output += 'hUS:   ' + str(hUS) + '\n' + \
                      'hDS:   ' + str(hDS) + '\n'

        elif self.sap_option == 'SapTrap':
            tag_type = self.tag_type
            hUS_seq = hUS
            hDS_seq = hDS
            gRNA_seq = DNA(self.site.gRNA.gRNA_sequence)
            sticky_ends = {'N': ['TTG', 'AAC', 'TGG', 'CAT', 'ACG', 'TAC'],
                           'C': ['TTG', 'AAC', 'TGG', 'CGC', 'GGT', 'TAC'],
                           'L': ['TTG', 'AAC', 'TGG', 'CGC', 'ACG', 'TAC'],
                           'X': ['TTG', 'AAC', 'TGG', 'CAT', 'GGT', 'TAC'],
                           'I': ['TTG', 'AAC', 'TGG', 'act', 'ttt', 'TAC'],
                           'S': ['TTG', 'AAC', 'TGG', 'acttac', 'tttcag', 'TAC']}
            hUS_T = sticky_ends[tag_type][2] + str(hUS_seq)
            hUS_B = sticky_ends[tag_type][3] + str(hUS_seq.reverse_complement())
            hDS_T = sticky_ends[tag_type][4] + str(hDS_seq)
            hDS_B = sticky_ends[tag_type][5] + str(hDS_seq.reverse_complement())
            sgRNA_T = sticky_ends[tag_type][0] + str(gRNA_seq)
            sgRNA_B = sticky_ends[tag_type][1] + str(gRNA_seq.reverse_complement())
            if tag_type == 'S':
                hUS_T += 'gta'
                hDS_B += 'ctg'

            if view == 'Alignment':
                output += 'Alignment of annealed oligos:\n'
                output += hUS_T + '\n'
                output += '   ' + str(DNA(hUS_B).reverse()) + '\n\n'
                output += hDS_T + '\n'
                output += '   ' + str(DNA(hDS_B).reverse()) + '\n\n'
                output += sgRNA_T + '\n'
                output += '   ' + str(DNA(sgRNA_B).reverse()) + '\n\n'

            if 'Oligos' in view:
                output += "Oligos in 5' to 3' direction\n\n"
                output += self.jobname + ' hUS_T\t' + hUS_T + '\n'
                output += self.jobname + ' hUS_B\t' + hUS_B + '\n'
                output += self.jobname + ' hDS_T\t' + hDS_T + '\n'
                output += self.jobname + ' hDS_B\t' + hDS_B + '\n'
                output += self.jobname + ' sgRNA_T\t' + sgRNA_T + '\n'
                output += self.jobname + ' sgRNA_B\t' + sgRNA_B + '\n\n'
            synthetic_DNA = 'gtgGCTCTTCg' + sticky_ends[tag_type][0] + str(gRNA_seq) + \
                            str(DNA(sticky_ends[tag_type][1]).reverse_complement()) + \
                            'cGAAGAGC' + \
                            'gtgGCTCTTCg' + sticky_ends[tag_type][2] + str(hUS_seq) + \
                            str(DNA(sticky_ends[tag_type][3]).reverse_complement()) + \
                            'cGAAGAGC' + \
                            'gtgGCTCTTCg' + sticky_ends[tag_type][4] + str(hDS_seq) + \
                            str(DNA(sticky_ends[tag_type][5]).reverse_complement()) + \
                            'cGAAGAGCgtg\n\n'

            if 'Synthetic' in view:
                if 'Annotated' not in view:
                    output += 'Synthetic DNA\n\n'
                    output += synthetic_DNA

            if 'Annotated' in view:
                gRNA_len = len(self.site.gRNA.gRNA_sequence)
                hUS_len = len(hUS)
                hDS_len = len(hDS)
                if self.tag_type == 'S':
                    hUS_len += 3
                    hDS_len += 3
                total_length = 78 + gRNA_len + hUS_len + hDS_len
                output += 'Annotated Synthetic DNA:\n\n'
                list_annotation = [['spacer', 3], ['SapI', 7], ['spacer', 1], ['sticky', 3], ['gRNA', gRNA_len],
                                   ['sticky', 3], ['spacer', 1], ['SapI', 7], ['spacer', 3], ['SapI', 7],
                                   ['spacer', 1], ['sticky', 3], ['hUS', hUS_len], ['sticky', 3], ['spacer', 1],
                                   ['SapI', 7], ['spacer', 3], ['SapI', 7], ['spacer', 1], ['sticky', 3],
                                   ['hDS', hDS_len], ['sticky', 3], ['spacer', 1], ['SapI', 7]]
                count = 1
                sapI_counter = 0
                sticky_counter = 0
                list_output_constructor = []
                for feature in list_annotation:
                    feature_location = str(count) + '..' + str(count + feature[1] - 1)
                    count += feature[1]
                    if feature[0] == 'spacer':
                        continue
                    elif feature[0] == 'SapI':
                        feature_label = 'SapI_site'
                        if sapI_counter > 0:
                            feature_label += '(' + str(sapI_counter) + ')'
                        feature_color = "#ff5375"
                        sapI_counter += 1
                    elif feature[0] == 'sticky':
                        feature_label = 'SapTrap overhang'
                        if sticky_counter > 0:
                            feature_label += '(' + str(sticky_counter) + ')'
                        feature_color = 'cyan'
                        sticky_counter += 1
                    elif feature[0] == 'gRNA':
                        feature_label = self.jobname + ' sgRNA'
                        feature_color = "#ff974a"
                    elif feature[0] == 'hUS':
                        feature_label = self.jobname + ' hUS'
                        feature_color = "#68a368"
                    elif feature[0] == 'hDS':
                        feature_label = self.jobname + ' hDS'
                        feature_color = "#68a368"
                    list_output_constructor.append([feature_location, feature_label, feature_color])
                annotated_file = 'LOCUS\t\t' + self.jobname + '\t\t\t' + str(total_length) + \
                                 ' bp\nCOMMENT     ApEinfo: methylated:1\nFEATURES             ' \
                                 'Location/Qualifiers\n'
                for item in list_output_constructor:
                    annotated_file += '     misc_feature    ' + item[0] + '\n                     /locus_tag="' + item[
                        1] + \
                                      '"\n                     /label="' + item[
                                          1] + '"\n                     /ApEinfo_label="' + item[1] + \
                                      '"\n                     /ApEinfo_fwdcolor="' + item[
                                          2] + '"\n                     /ApEinfo_revcolor="' + \
                                      item[
                                          2] + '"\n                     /ApEinfo_graphicformat="arrow_data {{0 1 2 0 0 -1} {} 0}' + \
                                      '\n                     width 5 offset 0"\n'
                annotated_file += 'ORIGIN\n' + synthetic_DNA + '//'
                output += annotated_file

        return output


def incrementer(initial_value, maximum, minimum=0, increment=1, circular=False):
    """ Increments through a range of integers(minimum, maximum) by increment(default = 1)
        If circular = 'No', will return minimum or maximum when asked to increment beyond range
        If circular = 'Yes', minimum - 1 = maximum, maximum + 1 = minimum
        Returns incremented value """
    return_value = None
    initial_sum = initial_value + increment
    if initial_sum in range(minimum, maximum + 1):
        return_value = initial_sum
    elif circular == True:
        true_increment = ((increment - (maximum + 1 - initial_value)) % (maximum + 1 - minimum))
        return_value = minimum + true_increment
    elif circular == False:
        if initial_sum > maximum:
            return_value = maximum
        elif initial_sum < minimum:
            return_value = minimum

    return return_value


def strip_DNA_sequence(sequence, length):
    """ input sequence(string) and length(integer)
        returns string with all non-ACGTacgt characters removed
        and \n characters inserted at intervals of length"""
    output = ''
    count = 0
    for base in sequence:
        if base in 'ACGTacgt':
            output += base
            count += 1
            if count == length:
                output += '\n'
                count = 0

    return output


def change_DNA_case(sequence, change):
    """ Input sequence(string) and change(u or l character
        if u, returns string in upper case, if l return string in lower case"""
    output = ''
    for base in sequence:
        if change == 'u':
            output += base.upper()
        elif change == 'l':
            output += base.lower()

    return output


def parse_input_sequence(sequence):
    """Input = user supplied string containing DNA sequence, \n's, and a single non-DNA character
       Output is a library with two items: 'sequence' is the DNA sequence, 'insertion_site' is the
       position in the sequence marked for insertion
       if more than one insertion site is marked, sequence will be empty string"""
    if sequence == 'CAT':
        sequence = ''
    # Temporary cheat code for testing
    elif '!' in sequence:
        sequence = STR.test_site
    elif '∆' in sequence:
        sequence = STR.test_deletion
    # End Temporary cheat code for testing
    clean_sequence = ''
    insertion_site = int(len(sequence) / 2)  # default value is middle of input sequence
    sequence_summary = {'deletion': False, 'change_nuclease': False}
    del_region = [0, 0]
    nonbase_characters = []
    for char in sequence:  # Loop will remove any characters that are not AGCTagct
        if char in 'AGCTagct':
            clean_sequence += char
        elif char == '<':
            del_region[0] = int(len(clean_sequence))
            insertion_site = del_region[0] + 1
            nonbase_characters.append(char)
            continue
        elif char == '>':
            del_region[1] = int(len(clean_sequence))
            sequence_summary['deletion'] = True
            sequence_summary['deletion_region'] = del_region
            nonbase_characters.append(char)
            continue
        elif char == '\n':
            continue
        elif char == '/':  # records insertion site as position of non-AGCTagct character
            insertion_site = int(len(clean_sequence))
            nonbase_characters.append(char)
        elif char == '#':
            sequence_summary['change_nuclease'] = True
            continue
        else:
            nonbase_characters.append(char)
    if nonbase_characters not in [[], ['<', '>'], ['/']]:  # re-prompt if 1 non-base characters were entered
        clean_sequence = 'Entry_Error'
    sequence_summary['sequence'] = clean_sequence
    sequence_summary['insert_site'] = insertion_site

    return sequence_summary


def translate(sequence):
    """ Input DNA sequence of upper-case characters. Output translated protein sequence."""
    translated_sequence = ''
    for base in sequence:
        if base.upper() in 'ACGT':
            translated_sequence += base.upper()
    if len(translated_sequence) % 3 == 0:  # if test trims sequence to multiple of 3
        coding_length = len(translated_sequence)
    else:
        coding_length = len(translated_sequence) - (len(translated_sequence) % 3)
    translation = ''  # initialize storage string and counter
    count = 0
    while count < coding_length:  # loops through codons, appends to translated string

        codon = translated_sequence[count: count + 3]
        translation += STR.dictGeneticCode[codon]
        count += 3

    return translation


def translate_upper_case(seq, frame, condensed=True):
    """ Input sequence of DNA characters and translation frame. Uppercase will be translated, lowercase discarded.
        Output translated amino acids. condensed leaves no spaces between amino acids, False spaces for DNA alignment"""
    parsed_seq = []
    seq_run = ''
    seq_translate = ''
    if seq[0].islower():
        prev_base_case = 'lower'
    elif seq[0] == '-':
        prev_base_case = 'lower'
    elif seq[0].isupper():
        prev_base_case = 'upper'
    for base in seq:
        if base.islower():
            base_case = 'lower'
        elif base == '-':
            base_case = 'lower'
        elif base.isupper():
            base_case = 'upper'
        if base_case == prev_base_case:
            seq_run = seq_run + base
        else:
            parsed_seq.append(seq_run)
            seq_run = '' + base
            if base.islower():
                prev_base_case = 'lower'
            elif base.isupper():
                prev_base_case = 'upper'
    parsed_seq.append(seq_run)

    for item in parsed_seq:
        if item.isupper():
            seq_translate += item
    protein_seq = translate(seq_translate[frame:])
    if condensed == True:
        return (protein_seq)
    else:
        formatted_seq = ''
        count = 0 - frame
        amino_acid = 0
        for base in seq:
            if base.isupper():
                count += 1
                if count % 3 == 1:
                    try:
                        formatted_seq += protein_seq[amino_acid]
                        amino_acid += 1
                    except IndexError:
                        count -= 1
                else:
                    formatted_seq += ' '
            else:
                formatted_seq += ' '
        return formatted_seq


def synonymous_codons(amino_acid):
    """ Argument is single letter amino acid code
        Returns list of codons for amino acid """
    try:
        codon_options = STR.dict_rev_genetic_code_elegans[amino_acid.upper()]
    except KeyError:
        codon_options = []
    return codon_options


def list_to_triplets(list, lead_offset=0):
    """ Input is a list of numbers (or other objects
        groups sets of 3 objects into sub-lists and returns list of these triplet lists
        lead offset will remove designated number of elements from the first list
        useful for binning DNA sequence positions into codon groups"""
    output_list = []
    triplet = []
    triplet_count = 3 - lead_offset
    for position in list:
        triplet.append(position)
        triplet_count += 1
        if triplet_count % 3 == 0:
            output_list.append(triplet)
            triplet_count = 0
            triplet = []
    if triplet != output_list[-1]:
        if triplet != []:
            output_list.append(triplet)
    return output_list


def get_n_bases(sequence, target_number, start=0, direction='right'):
    """ input DNA sequence, number n of bases to retrieve, and optional start position (default 0) and direction
        from start position, moving in direction (right or left), retrieves n DNA bases, ignores non-ACGTacgt char """
    increment = 1
    number = target_number
    if direction == 'left':
        increment *= -1
        max_number = len(sequence[:start])
    else:
        max_number = len(sequence[start:])
    if max_number < number:
        number = max_number
    i = start
    output_string = ''
    count = 0
    n_bases = 0
    while n_bases < number:
        try:
            if sequence[i] in 'acgtACGT':
                output_string += sequence[i]
                n_bases += 1
            i += increment
            count += 1
            if i < 0:
                break
        except IndexError:
            break
    if direction == 'left':
        output_string = output_string[::-1]
    output = {'sequence': output_string, 'count': count}

    return output


def count_bases(sequence):
    """ input is DNA sequence, returns number of acgtACGT characters in sequence """
    count = 0
    for base in sequence:
        if base in 'acgtACGT':
            count += 1
    return count


def difference_count(sequence_1, sequence_2):
    """ Input two strings(sequences)
        Counts the number of differences between 2 strings, return number of differences """
    string_length = len(sequence_1)
    if len(sequence_1) > len(sequence_2):
        string_length = len(sequence_2)
    difference_count = 0
    for i in range(string_length):
        if sequence_1[i] != sequence_2[i]:
            difference_count += 1
    if len(sequence_1) > len(sequence_2):
        difference_count += len(sequence_1) - len(sequence_2)
    elif len(sequence_2) > len(sequence_1):
        difference_count + - len(sequence_2) - len(sequence_1)

    return difference_count


def make_permutation_array(sequence, substitution_array):
    """ Input a sequence (DNA or Protein) and a substitution array library that contains sequences to be substituted for
        individual characters in the sequence.
        Returns a list of of all combinations of the substitutable values for all elements"""
    sequence_permutations = ['']
    for position in sequence:
        new_permutations = []
        for permutation in sequence_permutations:
            for substitution in substitution_array[position]:
                new_permutations.append(permutation + substitution)
        sequence_permutations = new_permutations

    return sequence_permutations


def strip_deletions(sequence):
    """ input a DNA sequence with possible - and space characters representing deleted bases relative to reference
        returns library: 'sequence' contains input sequence with - and spaces removed, 'deletions' contains
        a list of the position locations in original sequence that contained - or space. """
    stripped_seq = ''
    deletion_locations = []
    for i in range(len(sequence)):
        if sequence[i] in ['-', ' ']:
            deletion_locations.append(i)
        elif sequence[i] in 'ACGTacgt':
            stripped_seq += sequence[i]
    stripped_seq = DNA(stripped_seq)

    return {'sequence': stripped_seq, 'deletions': deletion_locations}


def restriction_site_ranges(enzyme, sequence):
    """ Input the name of an enzyme (key in STR.restriction_enzyme_list) and DNA sequence
        returns a list of ranges; each range is the range in sequence of a binding site for enzyme """
    rec_seq = STR.restriction_enzyme_list[enzyme]
    rec_length = len(rec_seq)
    stripped_seq = strip_deletions(sequence)['sequence']
    deletion_locations = strip_deletions(sequence)['deletions']
    binding_sites = []
    for start_site in stripped_seq.find_all(rec_seq):
        binding_sites.append([start_site, start_site + rec_length])
    for location in deletion_locations:
        for i in range(len(binding_sites)):
            for j in range(len(binding_sites[i])):
                if location < binding_sites[i][j]:
                    binding_sites[i][j] += 1
    # chew up leading -'s if present
    for binding_site in binding_sites:
        while sequence[binding_site[0]] == '-':
            binding_site[0] += 1
    binding_ranges = []
    for binding_site in binding_sites:
        binding_ranges.append(range(binding_site[0], binding_site[1]))

    return binding_ranges


def codon_positions(sequence, frame):
    """ Input DNA sequence and translation frame
        output a list of lists of 3 positions each. each list of 3 posiitons is a list of the locations in
        original sequence of a single codon. """
    codon_positions = []
    trim_count = frame
    for i in range(len(sequence)):
        if sequence[i] in 'ACGT':
            if trim_count > 0:
                trim_count -= 1
            else:
                codon_positions.append(i)
    codons = []
    codon = []
    for i in range(len(codon_positions)):
        if len(codon) == 3:
            codons.append(codon)
            codon = []
        codon.append(codon_positions[i])
    if len(codon) == 3:
        codons.append(codon)
    return codons


def max_len_by_position(array, position):
    max_length = 0
    for element in array:
        if len(element[position]) > max_length:
            max_length = len(element[position])
    return max_length


Cas9 = GuideSpecification('********!!!!!!!!!/\!!!NGG', 'Cas9', '#                /\  P')
Cpf1 = GuideSpecification('TTTN!!!!!!!!!!!!*******/*---\-', 'Cpf1', 'P     #             /    \ ')