# Copyright (C) 2018 Matthew L. Schwartz - all rights reserved
# Not for redistribution

from SapTrap_Builder import SapTrap_Reference as STR
from SapTrap_Builder import SapTrap_Base as STB
from SapTrap_Builder import SapTrap_Automatic as STA
import SapTrap_Text as STT
import tkinter as tk
from tkinter import messagebox


class Program_Controller:
    def __init__(self, master, app):
        self.master = master
        self.app = app
        self.display = app.main_display_panel
        self.settings = app.settings_panel
        self.strings = app.display_strings
        self.instructions = app.instruction_panel
        self.errors = app.error_handler
        self.progress = app.progress_panel
        self.site = STB.EditSite('CAT', STB.Cas9)
        self.app.edit_site = self.site
        self.section_loop = None
        self.autoprocessor = STA.AutoProcessor(self.site)

    def flow_control(self, identifier):
        """ Input an identifier string indicating the program step to execute
            Identifier is used to select code to execute for identified step """
        if identifier in STR.dict_program_control.values(): #later might protect entry into
            self.master.focus_force()
            self.change_step(identifier)
        elif identifier == 'Autoprocess':
            self.update_display_string_list('Autoprocess')
            self.autoprocess()
        else:
            if self.section_loop is not None:
                self.section_loop(identifier)
                self.update_display()

        return 0

    def change_step(self, identifier):
        """ Changes active step among dict_program_control steps when called. """
        if identifier == 'Sequence Entry':
                self.update_display_string_list('Sequence Entry')
                self.update_display()
                self.section_loop = None
        elif identifier == 'Select Translation Frame':
            self.update_display_string_list('Select Translation Frame')
            self.select_translation_frame_setup()
            self.section_loop = self.select_translation_frame
        elif identifier == 'Select Tag Type':
            self.update_display_string_list('Select Tag Type')
            self.section_loop = self.select_tag_type
        elif identifier == 'Set Insert Site':
            self.update_display_string_list('Set Insert Site')
            self.set_insert_site_setup()
            self.section_loop = self.set_insert_site
        elif identifier == 'Select gRNA':
            self.update_display_string_list('Select gRNA')
            self.select_gRNA_setup()
            self.section_loop = self.select_gRNA
        elif identifier == 'Mutate Repair Template':
            self.update_display_string_list('Mutate Repair Template')
            self.mutate_repair_template_setup()
            self.section_loop = self.mutate_repair_template
        elif identifier == 'Review Design':
            if self.final_entry_validation(self.site.final_reagents):
                self.update_display_string_list('Review Design')
                self.review_design_setup()
                self.review_design()
            self.section_loop = None
        elif identifier == 'Get Reagent Sequences':
            if self.final_entry_validation(self.site.final_reagents):
                self.update_display_string_list('Get Reagent Sequences')
                self.review_design_setup()
                self.get_reagent_sequences()
                self.section_loop = None

        return True


    def update_display_string_list(self, identifier):
        """ input is identifier string for current active step
            using identifier, function selects appropriate lists of string keywords for display elements to be added
            to or removed from the display. Calls add and remove functions from the GUI to change status of strings """
        add_list = []
        remove_list = []
        if identifier == 'Sequence Entry':
            add_list +=     []
            remove_list +=  ['base_translation', 'translation_differences', 'mutant_translation', 'pointer',
                             'mutated_sequence',
                             'repair_template', 'DNA_differences', 'sequence', 'gRNA', 'PAM', 'Enzyme_labels',
                             'final_display']
        elif identifier == 'Select Translation Frame':
            add_list +=     ['base_translation', 'sequence', 'Enzyme_labels']
            remove_list +=  ['repair_template', 'final_display']
        elif identifier ==   'Select Tag Type':
            add_list +=     []
            remove_list +=  ['final_display']
        elif identifier == 'Set Insert Site':
            add_list +=     ['base_translation', 'sequence', 'Enzyme_labels']
            remove_list +=  ['repair_template', 'final_display']
        elif identifier == 'Select gRNA':
            add_list +=     ['base_translation', 'sequence', 'gRNA', 'PAM', 'Enzyme_labels']
            remove_list +=  ['repair_template', 'final_display']
        elif identifier == 'Mutate Repair Template':
            add_list +=     ['base_translation', 'translation_differences', 'mutant_translation', 'pointer',
                             'mutated_sequence', 'DNA_differences', 'sequence', 'gRNA', 'PAM', 'Enzyme_labels']
            remove_list +=  ['repair_template', 'final_display']
        elif identifier == 'Review Design':
            add_list +=     ['base_translation', 'translation_differences', 'mutant_translation', 'DNA_differences',
                             'sequence', 'gRNA', 'PAM', 'repair_template', 'Enzyme_labels']
            remove_list +=  ['pointer', 'mutated_sequence', 'final_display']
        elif identifier == 'Get Reagent Sequences':
            add_list +=     ['final_display']
            remove_list +=  ['base_translation', 'translation_differences', 'mutant_translation', 'DNA_differences',
                             'sequence', 'gRNA', 'PAM', 'repair_template', 'Enzyme_labels', 'pointer',
                             'mutated_sequence']
        elif identifier == 'Autoprocess':
            add_list +=     ['base_translation', 'translation_differences', 'mutant_translation', 'DNA_differences',
                             'sequence', 'gRNA', 'PAM', 'repair_template', 'repair_template', 'Enzyme_labels']
            remove_list +=  ['pointer', 'mutated_sequence', 'final_display']

        self.strings.add_display_strings(add_list)
        self.strings.remove_display_strings(remove_list)

        return 0

    def update_display_strings(self, *args):
        """ Input display line keyword identifiers
            Function updates the display string stored in the GUI display object to match current state of
              edit_site or gRNA objects """
        for arg in args:
            if arg == 'base_translation':
                self.strings.string_library[arg].update(self.site.translate(self.site.sequence))
            elif arg == 'translation_differences':
                self.strings.string_library[arg].update(self.site.difference_string('translation'))
            elif arg == 'mutant_translation':
                self.strings.string_library[arg].update(self.site.translate(self.site.mutated_sequence))
            elif arg == 'pointer':
                self.strings.string_library[arg].update(self.site.set_pointer(self.site.mutation_site))
            elif arg == 'mutated_sequence':
                self.strings.string_library[arg].update(self.site.mutated_sequence)
            elif arg == 'repair_template':
                self.strings.string_library[arg].update(self.site.final_reagents.display_string,
                                                   pad_length=self.site.final_reagents.display_offset)
            elif arg == 'DNA_differences':
                self.strings.string_library[arg].update(self.site.difference_string('DNA'))
            elif arg == 'sequence':
                self.strings.string_library[arg].update(self.site.sequence)
            elif arg == 'gRNA':
                self.strings.string_library[arg].update(self.site.gRNA.display_sequence['sequence'],
                                                   pad_length=self.site.gRNA.start)
            elif arg == 'PAM':
                if self.site.gRNA:
                    pad_offset = 0
                    if self.site.gRNA.direction == 'antisense':
                        pad_offset = self.site.guide_specification.rc_PAM_offset
                    self.strings.string_library[arg].update(self.site.PAM_string(),
                                                      pad_length=self.site.gRNA.start - pad_offset)
            elif arg == 'Enzyme_labels':
                self.strings.label_enzymes()
            elif arg == 'flags':
                flags = self.site.collect_flags(self.app.current_step.get())
                self.strings.update_active_flags(flags)
                if self.app.current_step.get() == 8:
                    pad_length = 0
                else:
                    if self.site.gRNA is not None:
                        pad_length = self.site.gRNA.start
                    else:
                        pad_length = self.site.insert_site
                for flag in flags:
                    self.strings.string_library[flag].update(STT.flags[flag], pad_length=pad_length)
            elif arg == 'final_display':
                self.strings.string_library[arg].update(
                            self.site.final_reagents.print_oligos(self.settings.output_type.get()))
            elif arg == 'active':
                active_strings = []
                for string_key in self.strings.string_library.keys():
                    if self.strings.string_library[string_key].active == True:
                        active_strings.append(string_key)
                for string in active_strings:
                    self.update_display_strings(string)
                self.update_display_strings('flags')
        return 0

    def update_display(self):
        """ Refreshes display. if Insert site is displayed, transmits current insert_site and text (captital or
        lower-case) to display. """
        if self.site.show_insert == True:
            self.display.update_display(self.site.insert_site, self.site.insert_text())
        else:
            self.display.update_display()

        return 0

    def sequence_entry_setup(self):
        """ Prepares program for sequence entry step. Asks user if sure they want to delete active site, and
        returns to previous step with old site if they decline. """
        last_step = self.app.last_step.get()
        reset_site = True
        if self.site.sequence != 'CAT':
            reset_site = messagebox.askokcancel('Delete Current Site?',
                                                'Returning to Sequence Entry will delete current site!')
            if reset_site:
                self.site.reset('CAT', STB.Cas9)
                self.strings.reset()
                self.display.main_display.configure(state='normal')
                self.display.main_display.delete('1.0', tk.END)
                self.app.current_step.set(1)
                return 1

        return reset_site


    def sequence_entry(self):
        """ Parses input sequence in text box; forces user to reenter if entry invalid format """
        parsed_input = STB.parse_input_sequence(self.display.get())
        guide = STB.Cas9  # convert from string to variable when add back guide choice
        if parsed_input['change_nuclease']:
            guide = STB.Cpf1  # convert from string to variable when add back guide choice
        if parsed_input['sequence'] == 'Entry_Error':
            self.errors.error_message('Non-DNA Input')
            return False
        elif parsed_input['sequence'] == '':
            self.errors.error_message('No DNA Input')
            return False
        else:
            self.site.reset(parsed_input['sequence'], guide, parsed_input['insert_site'])
            self.apply_sap_option()
        if parsed_input['deletion'] == True:
            self.site.incorporate_large_deletion(parsed_input['deletion_region'])
            self.strings.add_display_strings(['mutated_sequence'])
            self.update_display_strings('mutated_sequence')

        return True

    def apply_sap_option(self):
        """ Function reads sap_option from sap_option optionMenu, and changes sapI site display and insert site
        display in GUI based on selection"""
        sap_option = self.settings.sap_selection.get()
        self.site.final_reagents.sap_option = sap_option
        if sap_option == 'Markerless No Insert':
            self.strings.remove_enzyme_highlighting(STR.trap_enzyme)
            self.site.show_insert = False
            if self.site.gRNA is not None:
                self.site.insert_site = self.site.gRNA.cutsite
        if sap_option == 'Markerless + Insert':
            self.strings.remove_enzyme_highlighting(STR.trap_enzyme)
            self.site.show_insert = True
        elif sap_option == 'SapTrap':
            self.site.show_insert = True
            self.strings.highlight_enzyme(STR.trap_enzyme, 'salmon')
        self.update_display_strings('active')
        self.update_display()

        return 0

    def apply_tag_type(self):
        """ Applies selected tag type from tag-type option menu to program and GUI. """
        self.site.final_reagents.tag_type = self.progress.tag_keys[self.progress.tag_type.get()]
        self.update_display_strings('DNA_differences', 'flags')
        self.update_display()

    def select_translation_frame_setup(self):
        """ Updates sequence and translation using current frame. """
        self.update_display_strings('sequence', 'base_translation', 'flags', 'Enzyme_labels')
        self.update_display()
        self.display.main_display.see(self.site.feature_locator('insert_site') + '+10c')
        self.display.main_display.see(self.site.feature_locator('insert_site') + '-10c')

        return 0

    def select_translation_frame(self, response):
        """ Inputs are arrow keys. changes translation frame of sequence. """
        #direction = self.app.user_response.get()
        if response in ['right', 'up']:
            self.site.change_frame(1)
        elif response in ['left', 'down']:
            self.site.change_frame(2)
        self.update_display_strings('base_translation', 'mutant_translation', 'Enzyme_labels')

        return True

    def select_tag_type(self, response):
        """ Input is up or down arrow key
            Sets option menu to previous or next option from list """
        current_tag_type = self.progress.tag_type.get()
        max_tag_index = len(self.progress.tag_types) - 1
        try:
            current_tag_index = self.progress.tag_types.index(current_tag_type)
        except:
            current_tag_index = -1
        if response == 'up':
            new_tag_index = STB.incrementer(current_tag_index, max_tag_index, increment=-1, circular=True)
            self.progress.tag_type.set(self.progress.tag_types[new_tag_index])
        elif response == 'down':
            new_tag_index = STB.incrementer(current_tag_index, max_tag_index, increment=1, circular=True)
            self.progress.tag_type.set(self.progress.tag_types[new_tag_index])

        return True


    def set_insert_site_setup(self):
        """ updates display_strings to show relevant information.
            scrolls to initial insert_site. """
        if self.settings.sap_selection.get() == 'Markerless No Insert':
            self.site.show_insert = False
        else:
            self.site.show_insert = True
        self.update_display_strings('flags')
        self.update_display()
        self.display.main_display.see(self.site.feature_locator('insert_site'))

        return 0


    def set_insert_site(self, response):
        """ reads user_response for right, left, RIGHT, LEFT, 'c', and 'r'
            Directionals move the insert_site, r resets to original, and c moves to cutsite of gRNA, if not None """
        initial_insert_site = self.site.insert_site
        if response == 'right':
            self.site.insert_site += 1
        elif response == 'left':
            self.site.insert_site -= 1
        elif response == 'RIGHT':
            self.site.insert_site += 9
        elif response == 'LEFT':
            self.site.insert_site -= 9
        elif response == 'c':
            if self.site.gRNA is not None:
                self.site.insert_site = self.site.gRNA.cutsite
        elif response == 'r':
            self.site.insert_site = self.site.original_insert_site
        if self.site.insert_site < 0:
            self.site.insert_site = 0
        if self.site.insert_site > len(self.site.sequence):
            self.site.insert_site =len(self.site.sequence)
        if self.site.final_reagents.sap_option == 'Markerless No Insert':
            self.site.insert_site = initial_insert_site
        self.display.main_display.see(self.site.feature_locator('insert_site') + '+10c')
        self.display.main_display.see(self.site.feature_locator('insert_site') + '-5c')
        self.update_display_strings('PAM', 'DNA_differences')

        return True

    def select_gRNA_setup(self):
        """ Finds gRNA binding site in guide list for which difference between cutsite and insert site is smallest
        sets initial guide to be that for which distance is smallest """
        if self.site.guide_list == {}:
            self.app.error_handler.error_message('No guides available')
        else:
            selection = 1
            for i in range(1, 1 + len(self.site.guide_list)):
                self.site.guide_list[i].distance = abs(self.site.guide_list[i].cutsite -
                                                            self.site.insert_site)
                if self.site.guide_list[i].distance < self.site.guide_list[selection].distance:
                    selection = i
            if self.site.gRNA is None:
                self.site.set_gRNA(selection)
            self.update_display_strings('gRNA', 'flags', 'PAM')
            self.update_display()
            self.display.main_display.see(self.site.feature_locator('guide') + '+10c')
            self.display.main_display.see(self.site.feature_locator('guide') + '-10c')

        return 0

    def select_gRNA(self, response):
        """ Reads user_response variable for left or right - loops through guides in guide_list"""
        if self.site.guide_list == {}:
            self.app.error_handler.error_message('No guides available')
        else:
            max_guide = len(self.site.guide_list)
            if max_guide is None:  # Need to test this that loop exits upon no guides found
                return False
            if response == 'right':
                self.site.set_gRNA(STB.incrementer(self.site.gRNA.index, max_guide, minimum=1, circular=True))
            if response == 'left':
                self.site.set_gRNA(STB.incrementer(self.site.gRNA.index, max_guide, minimum=1, increment=-1, circular=True))
            if self.site.final_reagents.sap_option == 'Markerless No Insert':
                self.site.insert_site = self.site.gRNA.cutsite
            self.update_display_strings('gRNA', 'flags', 'PAM', 'DNA_differences')
            self.display.main_display.see(self.site.feature_locator('guide') + '+20c')
            self.display.main_display.see(self.site.feature_locator('guide') + '-20c')

        return True

    def mutate_repair_template_setup(self):
        """ prepare mutator by copying current mutated sequence to a temporary hold (for resets)
            sets initial mutation site to gRNA cutsite, or insert of no gRNA selected"""
        try:
            self.site.final_reagents.tag_type = self.progress.tag_keys[self.progress.tag_type.get()]
        except KeyError:
            None
        self.site.temporary_sequence_hold = self.site.mutated_sequence
        if self.site.gRNA is not None:
            self.site.mutation_site = self.site.gRNA.cutsite
        else:
            self.site.mutation_site = self.site.insert_site
        self.update_display_strings('pointer', 'mutated_sequence', 'mutant_translation', 'flags')
        self.update_display()
        self.display.main_display.see(self.site.feature_locator('guide')  + '+10c')
        self.display.main_display.see(self.site.feature_locator('guide')  + '-10c')

        return 0

    def mutate_repair_template(self, response):
        """ recieves user input to specify navigation and mutation. Enacts cursor (active base) movements and
            mutation from a variety of keys. """
        if response in ['right', 'left', 'RIGHT', 'LEFT']:
            self.site.move_pointer(response)
            self.update_display_strings('pointer')
        elif response in ['Command-left', 'Command-right']:
            self.site.search_for_mutable_sites(response)
            self.update_display_strings('pointer')
        else:
            if response in ['r', 'z']:
                self.site.reset_mutator(response)
            elif response in STR.bases:
                self.site.base_mutator(response)
                if self.site.silent_site_search_region is not None:
                    self.site.reset_silent_sites(self.site.silent_site_search_region)
            elif response in STR.amino_acids:
                self.site.amino_acid_mutator(response)
                if self.site.silent_site_search_region is not None:
                    self.site.reset_silent_sites(self.site.silent_site_search_region)
            elif response in ['up', 'down', 'UP', 'DOWN']:
                self.site.scroll_mutator(response)
                if self.site.silent_site_search_region is not None:
                    self.site.reset_silent_sites(self.site.silent_site_search_region)
            elif response in ['Command-up', 'Command-down']:
                enzyme = self.site.silent_mutator(response)
                if enzyme in STR.restriction_enzyme_list.keys():
                    self.strings.reset_enzymes()
                    self.strings.highlight_enzyme(enzyme, 'cyan')
                    if self.site.final_reagents.sap_option == 'SapTrap':
                        self.strings.highlight_enzyme('SapI', 'salmon')
                    self.update_display_strings('mutated_sequence')
            self.update_display_strings('mutant_translation', 'mutated_sequence', 'DNA_differences',
                                        'translation_differences')
            self.strings.label_enzymes()
        self.display.main_display.see(self.site.feature_locator('mutation') + '+10c')
        self.display.main_display.see(self.site.feature_locator('mutation') + '-10c')
        return True

    def final_entry_validation(self, final_reagents):
        """ Checks data states and returns False if any data needed for steps 7 or 8 are missing.
            Make sure that none of the required fields / values are set to None or inappropriate values
            before determining homology arms for final preview or final calculation of reagents """
        final_reagents.sap_option = self.settings.sap_selection.get()
        if final_reagents.sap_option == 'Markerless No Insert':
            final_reagents.insert_site = self.site.gRNA.cutsite
        if final_reagents.sap_option == 'SapTrap':
            if final_reagents.tag_type is None:
                try:
                    final_reagents.tag_type = self.progress.tag_keys[self.progress.tag_type.get()]
                except:
                    self.errors.error_message('No Tag Type')
                    return False
            #elif final_reagents.tag_type is not None:
            #        final_reagents.tag_type = self.progress.tag_keys[self.progress.tag_type.get()]
        if self.site.gRNA == None:
            self.errors.error_message('No guide RNA')
            return False
        for user_input in [self.settings.hUS_length_entry, self.settings.hDS_length_entry,
                           self.settings.min_length_entry]:
            try:
                int(user_input.get('1.0', tk.END))
            except ValueError:
                self.errors.error_message('Non-integer input')
                return False
        final_reagents.hUS_target = int(self.settings.hUS_length_entry.get('1.0', tk.END))
        final_reagents.hDS_target = int(self.settings.hDS_length_entry.get('1.0', tk.END))
        final_reagents.min_true_length = int(self.settings.min_length_entry.get('1.0', tk.END))
        final_reagents.jobname = self.settings.jobname.get()

        return True

    def review_design_setup(self):
        """ Setup for review design - reset flags, retrieve homology arms, and evaluate the location of the gRNA
            binding site within the repair template. """
        self.site.final_reagents.reset_flags()
        self.site.final_reagents.get_homology_arms()
        self.site.final_reagents.evaluate_guide_binding_site()

    def review_design(self):
        """ Updates display to show repair template that will be produced and show flags regarding repair template"""
        self.update_display_strings('repair_template', 'flags')
        self.update_display()
        self.display.main_display.see(self.site.feature_locator('guide') + '+10c')
        self.display.main_display.see(self.site.feature_locator('guide') + '-10c')

        return True

    def get_reagent_sequences(self):
        """ prints final sequences of reagents and any flags / warnings about designs """
        self.site.final_reagents.print_oligos(self.settings.output_type.get())
        self.update_display_strings('flags', 'final_display')
        self.update_display()
        return True

    def autoprocess(self):
        self.autoprocessor.autoprocess_site()
        self.review_design_setup()
        self.update_display_strings('base_translation', 'translation_differences', 'mutant_translation',
                                    'mutated_sequence',
                                    'repair_template', 'DNA_differences', 'sequence', 'gRNA', 'PAM', 'Enzyme_labels',
                                    'final_display')
        self.update_display()
        self.app.change_active_step('Review Design')
